import warnings

import seaborn as sns
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import Voronoi


def plot_posterior_histogram(model, variables, number_samples=300): #TODO: fix code duplication

    # Get samples
    sample = model.get_sample(number_samples)
    post_sample = model.get_posterior_sample(number_samples)

    # Join samples
    sample["Mode"] = "Prior"
    post_sample["Mode"] = "Posterior"
    subsample = sample[variables + ["Mode"]]
    post_subsample = post_sample[variables + ["Mode"]]
    joint_subsample = subsample.append(post_subsample)

    # Plot posterior
    warnings.filterwarnings('ignore')
    g = sns.PairGrid(joint_subsample, hue="Mode")
    g = g.map_offdiag(sns.distplot)
    g = g.map_diag(sns.distplot)
    g = g.add_legend()
    warnings.filterwarnings('default')


def plot_posterior(model, variables, number_samples=1000):

    # Get samples
    sample = model.get_sample(number_samples)
    post_sample = model.get_posterior_sample(number_samples)

    # Join samples
    sample["Mode"] = "Prior"
    post_sample["Mode"] = "Posterior"
    subsample = sample[variables + ["Mode"]]
    post_subsample = post_sample[variables + ["Mode"]]
    joint_subsample = subsample.append(post_subsample)

    # Plot posterior
    warnings.filterwarnings('ignore')
    g = sns.PairGrid(joint_subsample, hue="Mode")
    g = g.map_offdiag(sns.kdeplot)
    g = g.map_diag(sns.kdeplot, lw=3, shade=True)
    g = g.add_legend()
    warnings.filterwarnings('default')


def plot_density(model, variables, number_samples=2000):
    sample = model.get_sample(number_samples)
    warnings.filterwarnings('ignore')
    g = sns.PairGrid(sample[variables])
    g = g.map_offdiag(sns.kdeplot)
    g = g.map_diag(sns.kdeplot, lw=3, shade=True)
    g = g.add_legend()
    warnings.filterwarnings('default')


def ensemble_histogram(sample_list, variable, weights, bins=30):
    num_samples = sum([len(s) for s in sample_list])
    num_resamples = [int(np.ceil(w*num_samples*2)) for w in weights]
    max_samples = max(num_resamples)
    hist_df = pd.DataFrame()
    for idx, s in enumerate(sample_list):
        num_remaining_samples = max_samples - num_resamples[idx]
        resampled_values = np.concatenate([s[variable].sample(num_resamples[idx], replace=True).values,
                                           np.array([np.nan]*num_remaining_samples)])
        hist_df["Model {}".format(idx)] = resampled_values
    hist_df.plot.hist(stacked=True, bins=bins)


def plot_particles(particles, var_name, var2_name=None, dim1=0, dim2=0, xlim=None, ylim=None, colors=None):
    if not var2_name:
        var2_name = var_name
    extracted_particles = np.transpose(np.array([[p.get_variable(name).value.detach().numpy().flatten()[dim] for p in particles]
                            for dim, name in [(dim1, var_name), (dim2, var2_name)]]))

    def voronoi_finite_polygons_2d(vor, radius=1000):
        """
        Reconstruct infinite voronoi regions in a 2D diagram to finite
        regions.

        Parameters
        ----------
        vor : Voronoi
            Input diagram
        radius : float, optional
            Distance to 'points at infinity'.

        Returns
        -------
        regions : list of tuples
            Indices of vertices in each revised Voronoi regions.
        vertices : list of tuples
            Coordinates for revised Voronoi vertices. Same as coordinates
            of input vertices, with 'points at infinity' appended to the
            end.

        """

        if vor.points.shape[1] != 2:
            raise ValueError("Requires 2D input")

        new_regions = []
        new_vertices = vor.vertices.tolist()

        center = vor.points.mean(axis=0)
        if radius is None:
            radius = vor.points.ptp().max()

        # Construct a map containing all ridges for a given point
        all_ridges = {}
        for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
            all_ridges.setdefault(p1, []).append((p2, v1, v2))
            all_ridges.setdefault(p2, []).append((p1, v1, v2))

        # Reconstruct infinite regions
        for p1, region in enumerate(vor.point_region):
            vertices = vor.regions[region]

            if all(v >= 0 for v in vertices):
                # finite region
                new_regions.append(vertices)
                continue

            # reconstruct a non-finite region
            ridges = all_ridges[p1]
            new_region = [v for v in vertices if v >= 0]

            for p2, v1, v2 in ridges:
                if v2 < 0:
                    v1, v2 = v2, v1
                if v1 >= 0:
                    # finite ridge: already in the region
                    continue

                # Compute the missing endpoint of an infinite ridge

                t = vor.points[p2] - vor.points[p1]  # tangent
                t /= np.linalg.norm(t)
                n = np.array([-t[1], t[0]])  # normal

                midpoint = vor.points[[p1, p2]].mean(axis=0)
                direction = np.sign(np.dot(midpoint - center, n)) * n
                far_point = vor.vertices[v2] + direction * radius

                new_region.append(len(new_vertices))
                new_vertices.append(far_point.tolist())

            # sort region counterclockwise
            vs = np.asarray([new_vertices[v] for v in new_region])
            c = vs.mean(axis=0)
            angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
            new_region = np.array(new_region)[np.argsort(angles)]

            # finish
            new_regions.append(new_region.tolist())

        return new_regions, np.asarray(new_vertices)

    vor = Voronoi(extracted_particles)

    # plot
    regions, vertices = voronoi_finite_polygons_2d(vor)

    # colorize
    if colors is None:
        for region in regions:
            polygon = vertices[region]
            plt.fill(*zip(*polygon), alpha=0.3)
    else:
        for idx, region in enumerate(regions):
            polygon = vertices[region]
            plt.fill(*zip(*polygon), alpha=0.3, color=colors[idx])

    plt.plot(extracted_particles[:, 0], extracted_particles[:, 1], 'ko')
    if xlim is None:
        xlim = (vor.min_bound[0] - 0.5, vor.max_bound[0] + 0.5)
    if ylim is None:
        ylim = (vor.min_bound[1] - 0.5, vor.max_bound[1] + 0.5)
    plt.xlim(xlim[0], xlim[1])
    plt.ylim(ylim[0], ylim[1])






import types
from copy import copy

import torch

from brancher.variables import var2link
from brancher.variables import Variable, PartialLink
from brancher.utilities import batch_meshgrid as torch_batch_meshgrid
from brancher.utilities import delta as torch_delta

class BrancherFunction(object):
    """
    Wrapper on backend functions (torch) for the user interface
    ----------
    """

    def __init__(self, fn, name="f_?"):
        self.fn = fn
        self.name = name
        if isinstance(fn, (torch.nn.Module, torch.nn.Sequential, torch.nn.ModuleList, torch.nn.ModuleDict,
                           torch.nn.Parameter, torch.nn.ParameterDict, torch.nn.ParameterList)): #TODO: all optimizable types in nn
            self.links = {fn}
        else:
            self.links = set()

    def _get_string(self, *args, **kwargs):
        return self.name + "(" + ", ".join([var2link(a).__str__()
                                            for n, a in enumerate(list(args) + list(kwargs.values()))]) + ")"

    def __call__(self, *args, **kwargs):
        link_args = [var2link(arg) for arg in args]
        link_kwargs = {name: var2link(arg) for name, arg in kwargs.items()}
        arg_vars = {var for link in link_args if isinstance(link, PartialLink) for var in link.vars}
        kwarg_vars = {var for _, link in link_kwargs.items() if isinstance(link, PartialLink) for var in link.vars}

        def fn(values):
            args = [x.fn(values) if isinstance(x, PartialLink) else x for x in link_args]
            kwargs = dict({(name, x.fn(values)) if isinstance(x, PartialLink) else (name, x)
                           for name, x in link_kwargs.items()})
            return self.fn(*args, **kwargs)

        return PartialLink(arg_vars.union(kwarg_vars), fn,
                           self.links, string=self._get_string(*args, **kwargs))

    @staticmethod
    def _is_var(self, arg):
        return isinstance(arg, (Variable, PartialLink))

# function set as a combination of torch.nn.functional and torch._C._VariableFunctions
# comprises most operations on tensors, including math, reshashing, broadcasting and nn functions

fn_set = {}

for name, fun in torch.nn.functional.__dict__.items():
    fn_set[name] = getattr(torch.torch.nn.functional, name)

for name in dir(torch._C._VariableFunctions):
    if name.startswith('_'):
        continue
    fn_set[name] = getattr(torch.torch._C._VariableFunctions, name)

is_backend_fn = lambda k, v: type(v) in [types.FunctionType, types.BuiltinFunctionType] and not k.startswith('_') #TODO: Work in progress
brancher_fns = {name: BrancherFunction(v, name) for name, v in fn_set.items() if is_backend_fn(name, v)}
globals().update(brancher_fns)

## Custom functions ##
batch_meshgrid = BrancherFunction(torch_batch_meshgrid)
delta = BrancherFunction(torch_delta)


## Add batch dimension to NN weights in PyTorch convolutions ##
def __batch_conv(conv_func, *args, **kwargs):
    args = list(args)
    inpt_in_args = 0
    if "input" in kwargs:
        in_data = kwargs.pop("input")
    else:
        in_data = args.pop(0)
        inpt_in_args += 1
    if "weight" in kwargs:
        weights = kwargs.pop("weight")
    else:
        weights = args.pop(1 - inpt_in_args)
    args = tuple(args)
    if "groups" in kwargs:
        raise ValueError("Group convolutions are currently not supported in Brancher.")
    in_channels = in_data.shape[1]
    out_channels = weights.shape[1]
    data_size = in_data.shape[2:]
    kernel_size = weights.shape[3:]
    batch_size = in_data.shape[0]
    in_data_t = in_data.view((1, batch_size*in_channels) + data_size)
    weights_t = weights.view((batch_size*out_channels, in_channels) + kernel_size)
    out_t = conv_func(in_data_t, weights_t, groups=batch_size, *args, **kwargs)
    out_size = out_t.shape[2:]
    out = out_t.view((batch_size, out_channels) + out_size)
    return out


def __batch_conv1d(*args, **kwargs):
    return __batch_conv(torch.nn.functional.conv1d, *args, **kwargs)


def __batch_conv2d(*args, **kwargs):
    return __batch_conv(torch.nn.functional.conv2d, *args, **kwargs)


def __batch_conv3d(*args, **kwargs):
    return __batch_conv(torch.nn.functional.conv3d, *args, **kwargs)


def __batch_conv_transpose1d(*args, **kwargs):
    return __batch_conv(torch.nn.functional.conv_transpose1d, *args, **kwargs)


def __batch_conv_transpose2d(*args, **kwargs):
    return __batch_conv(torch.nn.functional.conv_transpose2d, *args, **kwargs)


def __batch_conv_transpose3d(*args, **kwargs):
    return __batch_conv(torch.nn.functional.conv_transpose3d, *args, **kwargs)


conv1d = BrancherFunction(__batch_conv1d)
conv2d = BrancherFunction(__batch_conv2d)
conv3d = BrancherFunction(__batch_conv3d)

conv_transpose1d = BrancherFunction(__batch_conv_transpose1d)
conv_transpose2d = BrancherFunction(__batch_conv_transpose2d)
conv_transpose3d = BrancherFunction(__batch_conv_transpose3d)

## Lynear layers ##

def linear(input, weight, bias=None):
    return matmul(weight, input) + bias

## Reshape functions ##

def __reshape(input, shape):
    batch_size = input.shape[0]
    return torch.reshape(input, tuple([batch_size]) + shape)

reshape = BrancherFunction(__reshape)


"""
Variables
---------
Module description
"""

from abc import ABC, abstractmethod
import operator

import numpy as np

from brancher.variables import Variable, PartialLink
from brancher.standard_variables import MultivariateNormalVariable
from brancher.standard_variables import var2link
import brancher.functions as BF
from brancher.utilities import coerce_to_dtype

import torch


class StochasticProcess(ABC):

    @abstractmethod
    def __call__(self, query_points):
        pass


## Gaussian processes ##
class GaussianProcess(StochasticProcess):

    def __init__(self, mean_function, covariance_function, name):
        self.mean_function = mean_function
        self.covariance_function = covariance_function
        self.name = name

    def __call__(self, query_points):
        x = var2link(query_points)
        return MultivariateNormalVariable(loc=self.mean_function(x),
                                          covariance_matrix=self.covariance_function(x),
                                          name=self.name + "(" + query_points.name + ")")


class CovarianceFunction(ABC):

    def __init__(self, covariance):
        self.covariance = covariance

    def __call__(self, query_points1, query_points2=None):
        if not query_points2:
            query_points2 = query_points1
        query_grid = BF.batch_meshgrid(query_points1, query_points2)
        return self.covariance(query_grid[0], query_grid[1])

    def __add__(self, other):
        if isinstance(other, CovarianceFunction):
            return CovarianceFunction(covariance=lambda x, y: self.covariance(x, y) + other.covariance(x, y))
        elif isinstance(other, (int, float)):
            return CovarianceFunction(covariance=lambda x, y: self.covariance(x, y) + other)
        else:
            raise ValueError("Only covarianceFunctions and numbers can be summed to CovarianceFunctions")

    def __mul__(self, other):
        if isinstance(other, CovarianceFunction):
            return CovarianceFunction(covariance=lambda x, y: self.covariance(x, y)*other.covariance(x, y))
        if isinstance(other, (Variable, PartialLink)):
            other = var2link(other)
            return CovarianceFunction(covariance=lambda x, y: self.covariance(x, y) * other)
        elif isinstance(other, (int, float)):
            return CovarianceFunction(covariance=lambda x, y: self.covariance(x, y) * other)
        else:
            raise ValueError("Only covarianceFunctions and numbers can be multiplied with CovarianceFunctions")

    def __rmul__(self, other):
        return self.__mul__(other)


class SquaredExponentialCovariance(CovarianceFunction):

    def __init__(self, scale, jitter=0.):
        self.scale = var2link(scale)
        covariance = lambda x, y: BF.exp(-(x-y)**2/(2*scale**2)) + BF.delta(x, y)*jitter
        super().__init__(covariance=covariance)


class WhiteNoiseCovariance(CovarianceFunction):

    def __init__(self, magnitude, jitter=0.):
        self.magnitude = var2link(magnitude)
        covariance = lambda x, y: magnitude*BF.delta(x, y) + BF.delta(x, y)*jitter
        super().__init__(covariance=covariance)


class HarmonicCovariance(CovarianceFunction):

    def __init__(self, frequency, jitter=0.):
        self.frequency = var2link(frequency)
        covariance = lambda x, y: BF.cos(2*np.pi*self.frequency*(x - y)) + BF.delta(x, y)*jitter
        super().__init__(covariance=covariance)


class PeriodicCovariance(CovarianceFunction):

    def __init__(self, frequency, scale, jitter=0.):
        self.frequency = var2link(frequency)
        self.scale = var2link(scale)
        covariance = lambda x, y: BF.exp(-2*BF.sin(np.pi*self.frequency*(x - y))**2/scale**2) + BF.delta(x, y)*jitter
        super().__init__(covariance=covariance)


class ExponentialCovariance(CovarianceFunction):

    def __init__(self, scale, jitter=0.):
        self.scale = var2link(scale)
        covariance = lambda x, y: BF.exp(-BF.abs(x-y)/(scale)) + BF.delta(x, y)*jitter
        super().__init__(covariance=covariance)


class MeanFunction(ABC):

    def __init__(self, mean):
        self.mean = mean

    def __call__(self, query_points):
        return self.mean(query_points)

    def _apply_operator(self, other, op):
        """
        Args:

        Returns:
        """
        if isinstance(other, MeanFunction):
            return lambda x: op(self.mean(x), other.mean(x))
        elif isinstance(other, (int, float)):
            return lambda x: op(self.mean(x), other)

    def __add__(self, other):
        return self._apply_operator(other, operator.add)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self._apply_operator(other, operator.sub)

    def __rsub__(self, other):
        return -1*self.__sub__(other)

    def __mul__(self, other):
        return self._apply_operator(other, operator.mul)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._apply_operator(other, operator.truediv)

    def __rtruediv__(self, other):
        raise NotImplementedError

    def __pow__(self, other):
        return self._apply_operator(other, operator.pow)

    def __rpow__(self, other):
        raise NotImplementedError


class ConstantMean(MeanFunction):

    def __init__(self, value):
        value = value
        mean = lambda x: BF.delta(x, x)*value
        super().__init__(mean=mean)


"""
Distributions
---------
Module description
"""
from abc import ABC, abstractmethod
from collections.abc import Iterable
import copy

import numpy as np
from scipy.special import binom

import torch
from torch import distributions

from brancher.utilities import broadcast_and_squeeze_mixed
from brancher.utilities import broadcast_and_reshape_parent_value
from brancher.utilities import sum_data_dimensions
from brancher.utilities import is_discrete, is_tensor
from brancher.utilities import tensor_range
from brancher.utilities import map_iterable
from brancher.utilities import get_number_samples_and_datapoints

from brancher.config import device

#TODO: We need asserts checking for the right parameters

class Distribution(ABC):
    """
    Summary
    """
    def __init__(self):
        self.torchdist = None
        self.required_parameters = {}
        self.has_differentiable_samples = None
        self.is_finite = None
        self.is_discrete = None
        self.has_analytic_entropy = None
        self.has_analytic_mean = None
        self.has_analytic_var = None
        pass

    def check_parameters(self, **parameters):
        assert all([any([param in parameters for param in parameters_tuple]) if isinstance(parameters_tuple, tuple) else parameters_tuple in parameters
                    for parameters_tuple in self.required_parameters])

    @abstractmethod
    def _preprocess_parameters_for_log_prob(self, x, **parameters):
        pass

    @abstractmethod
    def _preprocess_parameters_for_sampling(self, **parameters):
        pass

    @abstractmethod
    def _postprocess_sample(self, sample, shape):
        pass

    @abstractmethod
    def _postprocess_log_prob(self, log_prob, number_samples, number_datapoints):
        pass

    def calculate_log_probability(self, x, **parameters):
        self.check_parameters(**parameters)
        x, parameters, number_samples, number_datapoints = self._preprocess_parameters_for_log_prob(x, **parameters)
        log_prob = self._calculate_log_probability(x, **parameters)
        log_prob = self._postprocess_log_prob(log_prob, number_samples, number_datapoints)
        return sum_data_dimensions(log_prob)

    def get_sample(self, differentiable=True, **parameters):
        self.check_parameters(**parameters)
        parameters, shape = self._preprocess_parameters_for_sampling(**parameters)
        pre_sample = self._get_sample(differentiable=differentiable, **parameters)
        sample = self._postprocess_sample(pre_sample, shape)
        return sample

    def get_mean(self, **parameters):
        self.check_parameters(**parameters)
        parameters, shape = self._preprocess_parameters_for_sampling(**parameters)
        pre_mean = self._get_mean(**parameters)
        mean = self._postprocess_sample(pre_mean, shape)
        return mean

    def get_variance(self, **parameters):
        self.check_parameters(**parameters)
        parameters, shape = self._preprocess_parameters_for_sampling(**parameters)
        pre_variance = self._get_variance(**parameters)
        variance = self._postprocess_sample(pre_variance, shape)
        return variance

    def get_entropy(self, **parameters):
        self.check_parameters(**parameters)
        parameters, shape = self._preprocess_parameters_for_sampling(**parameters)
        pre_entropy = self._get_entropy(**parameters)
        entropy = self._postprocess_sample(pre_entropy, shape)
        return entropy

    def _get_statistic(self, query, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        out_stat = query(self.torchdist(**parameters))
        return out_stat

    def _get_sample(self, differentiable, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        if self.has_differentiable_samples and differentiable:
            return self._get_statistic(lambda x: x.rsample(), **parameters)
        else:
            return self._get_statistic(lambda x: x.sample(), **parameters)

    def _get_mean(self, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        if self.has_analytic_mean:
            return self._get_statistic(lambda x: x.mean, **parameters)
        else:
            raise ValueError("The mean of the distribution cannot be computed analytically")

    def _get_variance(self, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        if self.has_analytic_var:
            return self._get_statistic(lambda x: x.variance, **parameters)
        raise ValueError("The variance of the distribution cannot be computed analytically")

    def _get_entropy(self, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        if self.has_analytic_entropy:
            return self._get_statistic(lambda x: x.entropy(), **parameters)
        else:
            raise ValueError("The entropy of the distribution cannot be computed analytically")

    def _calculate_log_probability(self, x, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        log_prob = self.torchdist(**parameters).log_prob(x)
        return log_prob


class ContinuousDistribution(Distribution):
    pass


class DiscreteDistribution(Distribution):
    pass


class UnivariateDistribution(Distribution):
    """
    Summary
    """

    def _preprocess_parameters_for_sampling(self, **parameters):
        parameters = broadcast_and_squeeze_mixed((), parameters)
        return parameters, None

    def _preprocess_parameters_for_log_prob(self, x, **parameters):
        tuple_x, parameters = broadcast_and_squeeze_mixed(tuple([x]), parameters)
        return tuple_x[0], parameters, None, None #TODO: add proper output here

    def _postprocess_sample(self, sample, shape=None):
        return sample

    def _postprocess_log_prob(self, log_prob, number_samples, number_datapoints):
        return log_prob


class ImplicitDistribution(Distribution):
    """
    Summary
    """

    def _preprocess_parameters_for_sampling(self, **parameters):
        return parameters, None

    def _preprocess_parameters_for_log_prob(self, x, **parameters):
        return x, parameters, None, None #TODO: add proper output here

    def _postprocess_sample(self, sample, shape=None):
        return sample

    def _calculate_log_probability(self, x, **parameters):
        return torch.tensor(np.zeros((1,1))).float().to(device) #TODO: Implement some checks here

    def _postprocess_log_prob(self, log_pro, number_samples, number_datapoints):
        return log_pro

    def _get_statistic(self, query, **parameters):
        raise NotImplemented


class VectorDistribution(Distribution):
    """
    Summary
    """
    def _preproces_vector_input(self, vector_input_dict, vector_names):
        shapes_dict = {par_name: list(par_value.shape)
                       for par_name, par_value in vector_input_dict.items()
                       if par_name in vector_names}
        reshaped_parameters = {par_name: par_value.contiguous().view(size=(shapes_dict[par_name][0], np.prod(
            shapes_dict[par_name][1:]))) if par_name in vector_names else par_value
                               for par_name, par_value in vector_input_dict.items()}
        tensor_shape = list(shapes_dict.values())[0][1:]
        return reshaped_parameters, tensor_shape

    def _preprocess_parameters_for_sampling(self, **parameters):
        number_samples, number_datapoints = get_number_samples_and_datapoints(parameters)
        parameters = map_iterable(lambda x: broadcast_and_reshape_parent_value(x, number_samples, number_datapoints), parameters)
        reshaped_parameters, tensor_shape = self._preproces_vector_input(parameters, self.vector_parameters)
        shape = tuple([number_samples, number_datapoints] + tensor_shape)
        return reshaped_parameters, shape

    def _preprocess_parameters_for_log_prob(self, x, **parameters):
        parameters_and_data = parameters
        parameters_and_data.update({"x_data": x})
        number_samples, number_datapoints = get_number_samples_and_datapoints(parameters_and_data)
        parameters_and_data = map_iterable(lambda y: broadcast_and_reshape_parent_value(y, number_samples, number_datapoints), parameters_and_data)
        vector_names = self.vector_parameters
        vector_names.add("x_data")
        reshaped_parameters_and_data, _ = self._preproces_vector_input(parameters_and_data, vector_names)
        x = reshaped_parameters_and_data.pop("x_data")
        return x, reshaped_parameters_and_data, number_samples, number_datapoints

    def _postprocess_sample(self, sample, shape):
        return sample.contiguous().view(size=shape)

    def _postprocess_log_prob(self, log_pro, number_samples, number_datapoints):
        return log_pro.contiguous().view(size=(number_samples, number_datapoints))


class CategoricalDistribution(VectorDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.one_hot_categorical.OneHotCategorical
        self.required_parameters = {("probs", "logits")}
        self.optional_parameters = {}
        self.vector_parameters = {"probs", "logits"}
        self.matrix_parameters = {}
        self.scalar_parameters = {}
        self.differentiable_samples = False
        self.finite = True
        self.discrete = True
        self.analytic_entropy = True
        self.analytic_mean = True
        self.analytic_var = True

    def _calculate_log_probability(self, x, **parameters):
        """
        One line description

        Parameters
        ----------

        Returns
        -------
        """
        vector_shape = parameters["probs"].shape if "probs" in parameters else parameters["logits"].shape
        if x.shape == vector_shape and tensor_range(x) == {0, 1}:
            dist = self.torchdist
        else:
            dist = distributions.categorical.Categorical

        log_prob = dist(**parameters).log_prob(x[:, 0])
        return log_prob


class MultivariateNormalDistribution(VectorDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = torch.distributions.multivariate_normal.MultivariateNormal
        self.required_parameters = {"loc", ("covariance_matrix", "precision_matrix", "scale_tril")}
        self.optional_parameters = {}
        self.vector_parameters = {"loc"}
        self.matrix_parameters = {}
        self.scalar_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class DirichletDistribution(VectorDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = torch.distributions.dirichlet.Dirichlet
        self.required_parameters = {"concentration"}
        self.optional_parameters = {}
        self.vector_parameters = {"concentration"}
        self.matrix_parameters = {}
        self.scalar_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class DeterministicDistribution(ImplicitDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.required_parameters = {"value"}
        self.has_differentiable_samples = True
        self.is_finite = True
        self.is_discrete = True
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True

    def _get_sample(self, differentiable, **parameters):
        """
        One line description

        Parameters
        ----------
        Returns
        -------
        """
        return parameters["value"]

    def _get_mean(self, **parameters):
        """
        One line description

        Parameters
        ----------
        Returns
        -------
        """
        return parameters["value"]

    def _get_variance(self, **parameters):
        """
        One line description

        Parameters
        ----------
        Returns
        -------
        """
        return torch.tensor(np.zeros((1, 1, 1))).float().to(device)

    def _get_entropy(self, **parameters):
        """
        One line description

        Parameters
        ----------
        Returns
        -------
        """
        return torch.tensor(np.zeros((1, 1, 1))).float().to(device)


class EmpiricalDistribution(ImplicitDistribution): #TODO: It needs to be reworked.
    """
    Summary
    """
    def __init__(self, batch_size, is_observed):
        super().__init__()
        self.required_parameters = {"dataset"}
        self.optional_parameters = {"indices", "weights"}
        self.batch_size = batch_size
        self.is_observed = is_observed
        self.has_differentiable_samples = False
        self.is_finite = True
        self.is_discrete = True
        self.has_analytic_entropy = True #TODO: this can be implemented
        self.has_analytic_mean = False
        self.has_analytic_var = False

    def _get_sample(self, differentiable, **parameters):
        """
        One line description

        Parameters
        ----------
        Returns
        -------
        Without replacement
        """
        dataset = parameters["dataset"]
        if "indices" not in parameters:
            if "weights" in parameters:
                weights = parameters["weights"]
                p = np.array(weights).astype("float64")
                p = p/np.sum(p)
            else:
                p = None
            if is_tensor(dataset):
                if self.is_observed:
                    dataset_size = dataset.shape[1]
                else:
                    dataset_size = dataset.shape[2]
            else:
                dataset_size = len(dataset)
            if dataset_size < self.batch_size:
                raise ValueError("It is impossible to have more samples than the size of the dataset without replacement")
            if is_discrete(dataset): #
                indices = np.random.choice(range(dataset_size), size=self.batch_size, replace=False, p=p)
            else:
                number_samples = dataset.shape[0]
                indices = [np.random.choice(range(dataset_size), size=self.batch_size, replace=False, p=p)
                           for _ in range(number_samples)]
        else:
            indices = parameters["indices"]

        if is_tensor(dataset):
            if isinstance(indices, list) and isinstance(indices[0], np.ndarray):
                if self.is_observed:
                    sample = torch.cat([dataset[n, k, :].unsqueeze(dim=0) for n, k in enumerate(indices)], dim=0)
                else:
                    sample = torch.cat([dataset[n, :, k, :].unsqueeze(dim=0) for n, k in enumerate(indices)], dim=0)

            elif isinstance(indices, list) and isinstance(indices[0], (int, np.int32, np.int64)):
                if self.is_observed:
                    sample = dataset[:, indices, :]
                else:
                    sample = dataset[:, :, indices, :]
            else:
                raise IndexError("The indices of an empirical variable should be either a list of integers or a list of arrays")
        else:
            sample = list(np.array(dataset)[indices])
        return sample

    def _get_entropy(self, **parameters):
        if "weights" in parameters:
            probs = parameters["weights"]
        else:
            if is_tensor(parameters["dataset"]):
                n = int(parameters["dataset"].shape[0])
            else:
                n = len(parameters["dataset"])
            probs = torch.Tensor(np.ones((n,))).float().to(device)
        return distributions.categorical.Categorical(probs=probs).entropy()


class NormalDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.normal.Normal
        self.required_parameters = {"loc", "scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class StudentTDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.studentT.StudentT
        self.required_parameters = {"df", "loc", "scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class UniformDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.uniform.Uniform
        self.required_parameters = {"low", "high"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class LogNormalDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.log_normal.LogNormal
        self.required_parameters = {"loc", "scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class CauchyDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.cauchy.Cauchy
        self.required_parameters = {"loc", "scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = False


class HalfCauchyDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.half_cauchy.HalfCauchy
        self.required_parameters = {"scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = False


class HalfNormalDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.half_normal.HalfNormal
        self.required_parameters = {"scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class Chi2Distribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.chi2.Chi2
        self.required_parameters = {"df"}
        self.optional_parameters = {}
        self.has_differentiable_samples = False
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = False
        self.has_analytic_mean = True
        self.has_analytic_var = True


class GumbelDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.gumbel.Gumbel
        self.required_parameters = {"loc", "scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class LaplaceDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.laplace.Laplace
        self.required_parameters = {"loc", "scale"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class ExponentialDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.exponential.Exponential
        self.required_parameters = {"rate"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class BetaDistribution(ContinuousDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.beta.Beta
        self.required_parameters = {"concentration1", "concentration0"}
        self.optional_parameters = {}
        self.has_differentiable_samples = True
        self.is_finite = False
        self.is_discrete = False
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class BinomialDistribution(UnivariateDistribution, DiscreteDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.binomial.Binomial
        self.required_parameters = {"total_count", ("probs", "logits")}
        self.optional_parameters = {}
        self.has_differentiable_samples = False
        self.is_finite = True
        self.is_discrete = True
        self.has_analytic_entropy = False
        self.has_analytic_mean = True
        self.has_analytic_var = True


class NegativeBinomialDistribution(UnivariateDistribution, DiscreteDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.negative_binomial.NegativeBinomial
        self.required_parameters = {"total_count", ("probs", "logits")}
        self.optional_parameters = {}
        self.has_differentiable_samples = False
        self.is_finite = True
        self.is_discrete = True
        self.has_analytic_entropy = False
        self.has_analytic_mean = True
        self.has_analytic_var = True


class BernoulliDistribution(UnivariateDistribution, DiscreteDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.bernoulli.Bernoulli
        self.required_parameters = {("probs", "logits")}
        self.optional_parameters = {}
        self.has_differentiable_samples = False
        self.is_finite = True
        self.is_discrete = True
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class GeometricDistribution(UnivariateDistribution, DiscreteDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.geometric.Geometric
        self.required_parameters = {("probs", "logits")}
        self.optional_parameters = {}
        self.has_differentiable_samples = False
        self.is_finite = True
        self.is_discrete = True
        self.has_analytic_entropy = True
        self.has_analytic_mean = True
        self.has_analytic_var = True


class PoissonDistribution(DiscreteDistribution, UnivariateDistribution):
    """
    Summary
    """
    def __init__(self):
        super().__init__()
        self.torchdist = distributions.poisson.Poisson
        self.required_parameters = {"rate"}
        self.optional_parameters = {}
        self.has_differentiable_samples = False
        self.is_finite = False
        self.is_discrete = True
        self.has_analytic_entropy = False
        self.has_analytic_mean = True
        self.has_analytic_var = True






"""
Optimizers
---------
Module description
"""
from abc import ABC, abstractmethod
from collections.abc import Iterable
import copy

import torch

from brancher.standard_variables import LinkConstructor
from brancher.modules import ParameterModule, EmptyModule
from brancher.variables import BrancherClass, Variable, ProbabilisticModel

from brancher.config import device


class ProbabilisticOptimizer(ABC):
    """
    Summary

    Parameters
    ----------
    optimizer : chainer optimizer
        Summary
    """
    def __init__(self, model, optimizer='SGD', **kwargs):
        assert isinstance(optimizer, str), 'Optimizer should be a name of available pytoch optimizers'
        self.link_set = set()
        self.module = None
        self.setup(model, optimizer, **kwargs) #TODO: add asserts for checking the params dictionary

    def _update_link_set(self, model):
        assert isinstance(model, BrancherClass) #TODO: add intuitive error message
        variable_set = model.flatten() if isinstance(model, ProbabilisticModel) else model.ancestors
        for var in variable_set:
            link = var.link if hasattr(var, 'link') else None
            if isinstance(link, (ParameterModule, LinkConstructor)):  # TODO: make sure that if user inputs nn.ModuleList, this works
                self.link_set.add(link)

    def add_variable2module(self, random_variable):
        """
        Summary
        """
        self._update_link_set(random_variable)
        for link in self.link_set:
            if isinstance(link, ParameterModule):
                self.module.append(link)
            elif isinstance(link, LinkConstructor):
                [self.module.append(l) for l in link]

    def setup(self, model, optimizer, **kwargs):
        self.module = EmptyModule()
        optimizer_class = getattr(torch.optim, optimizer)
        if isinstance(model, (Variable, ProbabilisticModel)):
            self.add_variable2module(model)
        elif isinstance(model, Iterable) and all([isinstance(submodel, (Variable, ProbabilisticModel))
                                                  for submodel in model]):
            [self.add_variable2module(submodel) for submodel in model]
        else:
            raise ValueError("Only brancher variables and iterable of variables can be added to a probabilistic optimizer")
        if list(self.module.parameters()):
            self.optimizer = optimizer_class(self.module.parameters(), **kwargs)
        else:
            self.optimizer = None
        self.module.to(device)

    def update(self):
        self.optimizer.step()

    def zero_grad(self):
        self.optimizer.zero_grad()



"""
Inference
---------
Module description
"""
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable

import numpy as np
from tqdm import tqdm

import torch

from brancher.optimizers import ProbabilisticOptimizer
from brancher.variables import Variable, ProbabilisticModel, Ensemble
from brancher.standard_variables import DeterministicVariable
from brancher.transformations import truncate_model
from brancher.variables import RootVariable
from brancher import gradient_estimators

from brancher.utilities import reassign_samples
from brancher.utilities import zip_dict
from brancher.utilities import sum_from_dim
from brancher.utilities import to_tensor


# def maximal_likelihood(random_variable, number_iterations, optimizer=chainer.optimizers.SGD(0.001)):
#     """
#     Summary
#
#     Parameters
#     ---------
#     random_variable : brancher.Variable
#     number_iterations : int
#     optimizer : chainer.optimizers
#     Summary
#     """
#     prob_optimizer = ProbabilisticOptimizer(optimizer) #TODO: This function is not up to date
#     prob_optimizer.setup(random_variable)
#     loss_list = []
#     for iteration in tqdm(range(number_iterations)):
#         loss = -F.sum(random_variable.calculate_log_probability({}))
#         prob_optimizer.chain.cleargrads()
#         loss.backward()
#         prob_optimizer.optimizer.update()
#         loss_list.append(loss.data)
#     return loss_list


def perform_inference(joint_model, number_iterations, number_samples = 1,
                      optimizer='Adam', input_values={},
                      inference_method=None,
                      posterior_model=None,
                      sampler_model=None,
                      pretraining_iterations=0,
                      **opt_params): #TODO: input values
    """
    Summary

    Parameters
    ---------
    """
    if isinstance(joint_model, Variable):
        joint_model = ProbabilisticModel([joint_model])
    if not inference_method:
        warnings.warn("The inference method was not specified, using the default reverse KL variational inference")
        inference_method = ReverseKL()
    if not posterior_model and joint_model.posterior_model is not None:
        posterior_model = joint_model.posterior_model
    else:
        posterior_model = inference_method.construct_posterior_model(joint_model)
    if not sampler_model: #TODO: clean up
        if not sampler_model:
            try:
                sampler_model = inference_method.sampler_model
            except AttributeError:
                try:
                    sampler_model = joint_model.posterior_sampler
                except AttributeError:
                    sampler_model = None

    joint_model.update_observed_submodel()

    def append_prob_optimizer(model, optimizer, **opt_params):
        prob_opt = ProbabilisticOptimizer(model, optimizer, **opt_params) # TODO: this should be better! handling models with no params
        if prob_opt.optimizer:
            optimizers_list.append(prob_opt)

    optimizers_list = []
    if inference_method.learnable_posterior:
        append_prob_optimizer(posterior_model, optimizer, **opt_params)
    if inference_method.learnable_model:
        append_prob_optimizer(joint_model, optimizer, **opt_params)
    if inference_method.learnable_sampler:
        append_prob_optimizer(sampler_model, optimizer, **opt_params)

    loss_list = []

    inference_method.check_model_compatibility(joint_model, posterior_model, sampler_model)

    for iteration in tqdm(range(number_iterations)):
        loss = inference_method.compute_loss(joint_model, posterior_model, sampler_model, number_samples)

        if torch.isfinite(loss.detach()).all().item():
            [opt.zero_grad() for opt in optimizers_list]
            loss.backward()
            inference_method.correct_gradient(joint_model, posterior_model, sampler_model, number_samples)
            optimizers_list[0].update()
            if iteration > pretraining_iterations:
                [opt.update() for opt in optimizers_list[1:]]
            loss_list.append(loss.cpu().detach().numpy().flatten())
        else:
            warnings.warn("Numerical error, skipping sample")
        loss_list.append(loss.cpu().detach().numpy())
    joint_model.diagnostics.update({"loss curve": np.array(loss_list)})

    inference_method.post_process(joint_model) #TODO: this could be implemented with a with block

    if joint_model.posterior_model is None and inference_method.learnable_posterior:
        joint_model.set_posterior_model(posterior_model)


class InferenceMethod(ABC):

    @abstractmethod
    def check_model_compatibility(self, joint_model, posterior_model, sampler_model):
        pass

    @abstractmethod
    def compute_loss(self, joint_model, posterior_model, sampler_model, number_samples, input_values):
        pass

    @abstractmethod
    def post_process(self, joint_model):
        pass

    def construct_posterior_model(self, joint_model):
        raise ValueError("Automatic construction of the posterior model is not currently implemented for this inference method. Set the posterior model manually")


class ReverseKL(InferenceMethod):

    def __init__(self, gradient_estimator=gradient_estimators.PathwiseDerivativeEstimator):
        self.learnable_posterior = True
        self.learnable_model = True
        self.needs_sampler = False
        self.learnable_sampler = False
        self.gradient_estimator = gradient_estimator

    def check_model_compatibility(self, joint_model, posterior_model, sampler_model):
        pass #TODO: Check differentiability of the model

    def compute_loss(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        loss = -joint_model.estimate_log_model_evidence(number_samples=number_samples,
                                                        method="ELBO", input_values=input_values,
                                                        for_gradient=True, gradient_estimator=self.gradient_estimator)
        return loss

    def correct_gradient(self, joint_model, posterior_model, sampler_model,
                         number_samples, input_values={}):
        pass

    def post_process(self, joint_model):
        pass


class WassersteinVariationalGradientDescent(InferenceMethod):

    def __init__(self, variational_samplers, particles,
                 cost_function=None,
                 deviation_statistics=None,
                 biased=False,
                 number_post_samples=20000,
                 gradient_estimator=gradient_estimators.PathwiseDerivativeEstimator):
        self.gradient_estimator = gradient_estimator
        self.learnable_posterior = True
        self.learnable_model = False #TODO: to implement later
        self.needs_sampler = True
        self.learnable_sampler = True
        self.biased = biased
        self.number_post_samples = number_post_samples
        if cost_function:
            self.cost_function = cost_function
        else:
            self.cost_function = lambda x, y: sum_from_dim((x - y)**2, dim_index=1)
        if deviation_statistics:
            self.deviation_statistics = deviation_statistics
        else:
            self.deviation_statistics = lambda lst: sum(lst)

        def model_statistics(dic):
            num_samples = list(dic.values())[0].shape[0]
            reassigned_particles = [reassign_samples(p._get_sample(num_samples), source_model=p, target_model=dic)
                                    for p in particles]

            statistics = [self.deviation_statistics([self.cost_function(value_pair[0].detach().numpy(),
                                                                        value_pair[1].detach().numpy())
                                                     for var, value_pair in zip_dict(dic, p).items()])
                          for p in reassigned_particles]
            return np.array(statistics).transpose()

        truncation_rules = [lambda a, idx=index: True if (idx == np.argmin(a)) else False
                            for index in range(len(particles))]

        self.sampler_model = [truncate_model(model=sampler,
                                             truncation_rule=rule,
                                             model_statistics=model_statistics)
                              for sampler, rule in zip(variational_samplers, truncation_rules)]

    def check_model_compatibility(self, joint_model, posterior_model, sampler_model):
        assert isinstance(sampler_model, Iterable) and all([isinstance(subsampler, (Variable, ProbabilisticModel))
                                                            for subsampler in sampler_model]), "The Wasserstein Variational GD method require a list of variables or probabilistic models as sampler"
        # TODO: Check differentiability of the model
        # TODO: check particles

    def compute_loss(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        sampler_loss = sum([-joint_model.estimate_log_model_evidence(number_samples=number_samples, posterior_model=subsampler,
                                                                     method="ELBO", input_values=input_values,
                                                                     for_gradient=True, gradient_estimator=self.gradient_estimator)
                            for subsampler in sampler_model])
        particle_loss = self.get_particle_loss(joint_model, posterior_model, sampler_model, number_samples,
                                               input_values)
        return sampler_loss + particle_loss #10

    def get_particle_loss(self, joint_model, particle_list, sampler_model, number_samples, input_values):
        samples_list = [sampler._get_sample(number_samples, input_values=input_values, max_itr=1)
                        for sampler in sampler_model]
        if self.biased:
            importance_weights = [1./number_samples for _ in sampler_model]
        else:
            importance_weights = [joint_model.get_importance_weights(q_samples=samples,
                                                                     q_model=sampler,
                                                                     for_gradient=False).flatten()
                                  for samples, sampler in zip(samples_list, sampler_model)]
        reassigned_samples_list = [reassign_samples(samples, source_model=sampler, target_model=particle)
                                   for samples, sampler, particle in zip(samples_list, sampler_model, particle_list)]
        pair_list = [zip_dict(particle._get_sample(1), samples)
                     for particle, samples in zip(particle_list, reassigned_samples_list)]

        particle_loss = sum([torch.sum(to_tensor(w)*self.deviation_statistics([self.cost_function(value_pair[0], value_pair[1].detach()) #TODO: numpy()
                                                                for var, value_pair in particle.items()]))
                             for particle, w in zip(pair_list, importance_weights)])
        return particle_loss

    def correct_gradient(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        pass

    def post_process(self, joint_model):
        sample_list = [sampler._get_sample(self.number_post_samples, max_itr=1)
                        for sampler in self.sampler_model]
        log_weights = []
        for sampler, s in zip(self.sampler_model, sample_list):
            _, logZ = joint_model.get_importance_weights(q_samples=s,
                                                         q_model=sampler,
                                                         for_gradient=False,
                                                         give_normalization=True)
            log_weights.append(logZ)
        log_weights = np.array(log_weights)
        alpha = np.max(log_weights)
        un_weights = np.exp(log_weights - alpha)
        self.weights = un_weights/np.sum(un_weights)
        #joint_model.set_posterior_model(Ensemble(self.sampler_model, self.weights)) #TODO: Work in progress


class MaximumLikelihood(InferenceMethod):

    def __init__(self):
        self.learnable_posterior = False
        self.learnable_model = True
        self.needs_sampler = False
        self.learnable_sampler = False

    def construct_posterior_model(self, joint_model):
        return None

    def check_model_compatibility(self, joint_model, posterior_model, sampler_model):
        # TODO: Check differentiability of the model
        pass

    def compute_loss(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        empirical_samples = joint_model.observed_submodel._get_sample(1, observed=True)
        loss = -joint_model.calculate_log_probability(empirical_samples, for_gradient=True)
        return loss.sum()

    def correct_gradient(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        pass

    def post_process(self, joint_model):
        pass


class MAP(InferenceMethod):

    def __init__(self):
        self.learnable_posterior = True
        self.learnable_model = True
        self.needs_sampler = False
        self.learnable_sampler = False

    def construct_posterior_model(self, joint_model):
        test_sample = joint_model._get_sample(1)
        posterior_model = ProbabilisticModel([DeterministicVariable(value[0, 0, :], variable.name, learnable=True)
                                              for variable, value in test_sample.items()
                                              if (not variable.is_observed) and not isinstance(variable, (DeterministicVariable, RootVariable))])
        return posterior_model

    def check_model_compatibility(self, joint_model, posterior_model, sampler_model):
        # TODO: Check differentiability of the model
        assert all([isinstance(var, (RootVariable, DeterministicVariable)) for var in posterior_model.flatten()])

    def compute_loss(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        empirical_samples = joint_model.observed_submodel._get_sample(1, observed=True)
        variable_values = reassign_samples(posterior_model._get_sample(1), source_model=posterior_model,
                                           target_model=joint_model)
        variable_values.update(empirical_samples)
        loss = -joint_model.calculate_log_probability(variable_values, for_gradient=True)
        return loss.sum()

    def correct_gradient(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        pass

    def post_process(self, joint_model):
        pass


class SteinVariationalGradientDescent(InferenceMethod): #TODO: work in progress

    def __init__(self):
        self.learnable_posterior = True
        self.learnable_model = False  # TODO: to implement later
        self.needs_sampler = False
        self.learnable_sampler = False
        self.deviation = lambda x, y: np.sum(((to_tensor(x) - to_tensor(y))**2).detach().numpy())
        self.kernel = lambda d, bw: np.exp(-d/(2*bw))
        self.bandwidth = 0.01

    def check_model_compatibility(self, joint_model, posterior_model, sampler_model):
        # TODO: Check differentiability of the model
        # TODO; check particles
        pass

    def compute_loss(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        empirical_samples = joint_model.observed_submodel._get_sample(1, observed=True)
        particle_samples = [reassign_samples(particle._get_sample(1), source_model=particle, target_model=joint_model)
                            for particle in posterior_model]
        [sample.update(empirical_samples) for sample in particle_samples]
        loss = sum([-joint_model.calculate_log_probability(sample, for_gradient=True)
                    for sample in particle_samples])
        return loss

    def correct_gradient(self, joint_model, posterior_model, sampler_model, number_samples, input_values={}):
        self.update_bandwidth(posterior_model)
        gradients = [[variable.value.grad for variable in particle.flatten()] for particle in posterior_model]
        kernel_matrix = [[self.kernel(sum([self.deviation(variable1.value, variable2.value)
                                          for variable1, variable2 in zip(particle1.flatten(), particle2.flatten())]),
                                      bw=self.bandwidth)
                          for particle1 in posterior_model] for particle2 in posterior_model]
        interaction_matrix = [[[-(variable1.value - variable2.value)*kernel_matrix[particle_index1][particle_index2]/self.bandwidth
                               for variable1, variable2 in zip(particle1.flatten(), particle2.flatten())]
                              for particle_index1, particle1 in enumerate(posterior_model)]
                              for particle_index2, particle2 in enumerate(posterior_model)]
        for particle_index, particle in enumerate(posterior_model):
            for variable_index, variable in enumerate(particle.flatten()):
                variable.value.grad = sum([kernel_matrix[particle_index][other_index]*other_gradient[variable_index] + interaction_matrix[particle_index][other_index][variable_index]
                                           for other_index, other_gradient in enumerate(gradients)])

    def update_bandwidth(self, posterior_model):
        distances = [np.sqrt(sum([self.deviation(variable1.value, variable2.value)
                                  for variable1, variable2 in zip(particle1.flatten(), particle2.flatten())]))
                     for particle1 in posterior_model
                     for particle2 in posterior_model
                     if particle1 is not particle2]
        bw = 2*np.median(distances)**2/np.log(len(posterior_model))
        self.bandwidth = bw

    def post_process(self, joint_model):
        pass





"""
Variables
---------
Module description
"""
from abc import ABC, abstractmethod
import operator
import numbers
import collections
from collections.abc import Iterable

from brancher.modules import ParameterModule

import numpy as np
import pandas as pd
import torch

import warnings

from brancher import distributions
from brancher import gradient_estimators

from brancher.utilities import join_dicts_list, join_sets_list
from brancher.utilities import flatten_list
from brancher.utilities import partial_broadcast
from brancher.utilities import coerce_to_dtype
from brancher.utilities import broadcast_and_reshape_parent_value
from brancher.utilities import split_dict
from brancher.utilities import reformat_sampler_input
from brancher.utilities import tile_parameter
from brancher.utilities import get_model_mapping
from brancher.utilities import reassign_samples
from brancher.utilities import is_discrete, is_tensor, contains_tensors
from brancher.utilities import concatenate_samples
from brancher.utilities import get_number_samples_and_datapoints
from brancher.utilities import map_iterable
from brancher.utilities import sum_from_dim

from brancher.pandas_interface import reformat_sample_to_pandas
from brancher.pandas_interface import reformat_model_summary
from brancher.pandas_interface import pandas_frame2dict
from brancher.pandas_interface import pandas_frame2value

from brancher.config import device


class BrancherClass(ABC): #This requires some refactoring, it should become a superclass of Variables and probabilistic models and Partial Links should be out
    """
    BrancherClass is the abstract superclass of all Brancher variables and models.
    """
    @abstractmethod
    def _flatten(self):
        """
        Abstract method. It returs a list of all the variables contained in the model.
        """
        pass

    @abstractmethod
    def _get_statistic(self, query, input_values):
        """
        Abstract method.
        """
        pass

    def flatten(self):
        return set(self._flatten())

    def get_variable(self, var_name):
        """
        It returns the variable in the model with the requested name.

        Args:
            var_name: String. Name  of the requested variable.

        Returns:
            torch.Tensor.

        """
        flat_list = self._flatten()
        try:
            return {var.name: var for var in flat_list}[var_name]
        except ValueError:
            raise ValueError("The variable {} is not present in the model".format(var_name))

    def _get_mean(self, input_values={}):
        return self._get_statistic(query=lambda dist, parameters: dist.get_mean(**parameters), input_values=input_values)

    def _get_variance(self, input_values={}):
        return self._get_statistic(query=lambda dist, parameters: dist.get_variance(**parameters), input_values=input_values)

    #def _get_entropy(self, input_values={}):
    #    if isinstance(self, ProbabilisticModel) or self.distribution.has_analytic_entropy:
    #        entropy_array = self._get_statistic(query=lambda dist, parameters: dist.get_entropy(**parameters),
    #                                            input_values=input_values)
    #        if isinstance(self, ProbabilisticModel):
    #            return sum([sum_from_dim(var_ent, 2) for var_ent in entropy_array.values()])
    #        else:
    #            return sum_from_dim(entropy_array, 2)
    #    else:
    #        return -self.calculate_log_probability(self, input_values, include_parents=False)


class Variable(BrancherClass):
    """
    Variable is the abstract superclass of deterministic and random variables. Variables are the building blocks of
    all probabilistic models in Brancher.
    """
    @abstractmethod
    def calculate_log_probability(self, values, reevaluate, include_parents):
        """
        Abstract method. It returns the log probability of the values given the model.

        Args:
            values: Dictionary(brancher.Variable: chainer.Variable). A dictionary having the brancher.variables of the
            model as keys and chainer.Variables as values. This dictionary has to provide values for all variables of
            the model except for the deterministic variables.

            reevaluate: Bool. If false it returns the output of the latest call. It avoid unnecessary computations when
            multiple children variables ask for the log probability of the same paternt variable.

        Returns:
            torch.Tensor. the log probability of the input values given the model.

        """
        pass

    @abstractmethod
    def _get_sample(self, number_samples, resample, observed, input_values, differentiable):
        """
        Abstract private method. It returns samples from the joint distribution specified by the model. If an input is provided
        it only samples the variables that are not contained in the input.

        Args:
            number_samples: Int.

            resample: Bool. If true it returns its previously stored sampled. It is used when multiple children variables
            ask for a sample to the same parent. In this case the resample variable is False since the children should be
            fed with the same value of the parent.

            observed: Bool. It specifies whether the samples should be interpreted frequentistically as samples from the
            observations of as Bayesian samples from the prior model. The first batch dimension is reserved to Bayesian
            samples while the second batch dimension is reserved to observation samples. Theerefore, this boolean
            changes the shape of the resulting sampled array.

            input_values: Dictionary(brancher.Variable, chainer.Variable). A dictionary having the brancher.variables of the
            model as keys and chainer.Variables as values. This dictionary has to provide values for all variables of
            the model that do not need to be sampled. Using an input allows to use a probabilistic model as a random
            function.

        Returns:
            Dictionary(brancher.Variable: torch.Tensor). A dictionary of samples from all the variables of the model

        """
        pass

    def _get_entropy(self, input_values={}):
        if self.distribution.has_analytic_entropy:
            entropy_array = self._get_statistic(query=lambda dist, parameters: dist.get_entropy(**parameters),
                                                input_values=input_values)
            return sum_from_dim(entropy_array, 2)
        else:
            return -self.calculate_log_probability(input_values, include_parents=False)

    def get_sample(self, number_samples, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=number_samples)
        raw_sample = {self: self._get_sample(number_samples, resample=False,
                                             observed=self.is_observed,
                                             differentiable=False,
                                             input_values=reformatted_input_values)[self]}
        sample = reformat_sample_to_pandas(raw_sample)
        self.reset()
        return sample

    def get_mean(self, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=1)
        raw_mean = {self: self._get_mean(reformatted_input_values)}
        mean = reformat_sample_to_pandas(raw_mean, number_samples=1)
        return mean

    def get_variance(self, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=1)
        raw_variance = {self: self._get_variance(reformatted_input_values)}
        variance = reformat_sample_to_pandas(raw_variance, number_samples=1)
        return variance

    def get_entropy(self, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=1)
        raw_ent = {self: self._get_entropy(reformatted_input_values)}
        ent = reformat_sample_to_pandas(raw_ent, number_samples=1)
        return ent

    @abstractmethod
    def reset(self):
        """
        Abstract method. It recursively resets the self._evalueted and self._current_value attributes of the variable and
        all downstream variables. It is used after sampling and evaluating the log probability of a model.

        Args: None.

        Returns: None.
        """
        pass

    @property
    @abstractmethod
    def is_observed(self):
        """
        Abstract property method. It returns True if the variable is observed and False otherwise.

        Args: None.

        Returns: Bool.
        """
        pass

    def __str__(self):
        """
        Method.

        Args: None

        Returns: String
        """
        return self.name

    def _apply_operator(self, other, op):
        """
        Method. It is used for performing symbolic operations between variables. It always returns a partialLink object
        that define a mathematical operation between variables. The vars attribute of the link is the set of variables
        that are used in the operation. The fn attribute is a lambda that specify the operation as a functions between the
        values of the variables in vars and a numeric output. This is required for defining the forward pass of the model.

        Args:
            other: PartialLink, RandomVariable, numeric or np.array.

            op: Binary operator.

        Returns: PartialLink
        """
        return var2link(self)._apply_operator(other, op)

    def __neg__(self):
        return -1*self

    def __add__(self, other):
        return self._apply_operator(other, operator.add)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self._apply_operator(other, operator.sub)

    def __rsub__(self, other):
        return -1*self.__sub__(other)

    def __mul__(self, other):
        return self._apply_operator(other, operator.mul)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._apply_operator(other, operator.truediv)

    def __rtruediv__(self, other):
        return self.__truediv__(other) ** (-1)

    def __pow__(self, other):
        return self._apply_operator(other, operator.pow)

    def __rpow__(self, other):
        raise NotImplementedError

    def __getitem__(self, key):
        if isinstance(key, str):
            variable_slice = key
        elif isinstance(key, Iterable):
            variable_slice = (slice(None, None, None), *key)
        else:
            variable_slice = (slice(None, None, None), key)
        vars = {self}
        fn = lambda values: values[self][variable_slice]
        links = set()
        return PartialLink(vars=vars, fn=fn, links=links)

    def shape(self):
        vars = {self}
        fn = lambda values: values[self].shape
        links = set()
        return PartialLink(vars=vars, fn=fn, links=links)


class RootVariable(Variable):
    """
    Deterministic variables are a subclass of random variables that always return the same value. The hyper-parameters of
    a probabilistic model are usually encoded as DeterministicVariables. When the user input a parameter as a Numeric value or
    an array, Brancher created a RootVariable that store its value.

    Parameters
    ----------
    data : torch.Tensor, numeric, or np.ndarray. The value of the variable. It gets stored in the self.value
    attribute.

    name : String. The name of the variable.

    learnable : Bool. This boolean value specify if the value of the RootVariable can be updated during traning.

    """
    def __init__(self, data, name, learnable=False, is_observed=False):
        self.name = name
        self.distribution = distributions.DeterministicDistribution()
        self._evaluated = False
        self._observed = is_observed
        self.parents = set()
        self.ancestors = set()
        self._type = "Deterministic"
        self.learnable = learnable
        self.link = None
        self._value = coerce_to_dtype(data, is_observed)
        if self.learnable:
            if not is_discrete(data):
                self._value = torch.nn.Parameter(coerce_to_dtype(data, is_observed), requires_grad=True)
                self.link = ParameterModule(self._value) # add to optimizer; opt checks links
            else:
                self.learnable = False
                warnings.warn('Currently discrete parameters are not learnable. Learnable set to False')

    def calculate_log_probability(self, values, reevaluate=True, for_gradient=False, normalized=True, include_parents=False):
        """
        Method. It returns the log probability of the values given the model. This value is always 0 since the probability
        of a deterministic variable having its value is always 1.

        Args:
            values: Dictionary(brancher.Variable: chainer.Variable). A dictionary having the brancher.variables of the
            model as keys and chainer.Variables as values. This dictionary has to provide values for all variables of
            the model except for the deterministic variables.

            reevaluate: Bool. If false it returns the output of the latest call. It avoid unnecessary computations when
            multiple children variables ask for the log probability of the same paternt variable.

        Returns:
            torch.Tensor. The log probability of the input values given the model.
        """

        return torch.Tensor(np.zeros((1, 1))).float().to(device)

    @property
    def value(self):
        if self.learnable:
            return self.link()
        return self._value

    @property
    def is_observed(self):
        return self._observed

    def _get_statistic(self, query, input_values):
        parameters_dict = {"value": self.value}
        statistic = query(self.distribution, parameters_dict)
        return statistic

    def _get_sample(self, number_samples, resample=False, observed=False, input_values={}, differentiable=True):
        if self in input_values:
            value = input_values[self]
        else:
            value = self.value
        if not is_discrete(value):
            return {self: tile_parameter(value, number_samples=number_samples)}
        else:
            return {self: value}

    def reset(self, recursive=False):
        pass

    def _flatten(self):
        return []


class RandomVariable(Variable):
    """
    Random variables are the main building blocks of probabilistic models.

    Parameters
    ----------
    distribution : brancher.Distribution
        The probability distribution of the random variable.
    name : str
        The name of the random variable.
    parents : tuple of brancher variables
        A tuple of brancher.Variables that are parents of the random variable.
    link : callable
        A function Dictionary(brancher.variable: torch.tensor) -> Dictionary(str: torch.tensor) that maps the values of all the parents
        to a dictionary of parameters of the probability distribution. It can also contains learnable layers and parameters.
    """
    def __init__(self, distribution, name, parents, link):
        self.name = name
        self.distribution = distribution
        self.link = link
        self.parents = parents
        self.ancestors = None
        self._type = "Random"
        self.samples = None

        self._evaluated = False
        self._observed = False # RandomVariable: observed value + link
        self._observed_value = None # need this?
        self.dataset = None
        self.has_random_dataset = False
        self.has_observed_value = False

    @property
    def value(self):
        """
        Method. It returns the value of the deterministic variable.

        Args:
            None

        Returns:
            torch.Tensor. The value of the deterministic variable.
        """
        if self._observed:
            return self._observed_value
        else:
            raise AttributeError('RandomVariable has to be observed to receive value.')

    @property
    def is_observed(self):
        return self._observed

    def _apply_link(self, parents_values):
        number_samples, number_datapoints = get_number_samples_and_datapoints(parents_values)
        cont_values, discrete_values = split_dict(parents_values,
                                                  condition=lambda key, val: not is_discrete(val) or contains_tensors(val))
        reshaped_dict = discrete_values
        if cont_values:
            reshaped_dict.update(map_iterable(lambda x: broadcast_and_reshape_parent_value(x, number_samples, number_datapoints),
                                              cont_values, recursive=True))
        reshaped_output = self.link(reshaped_dict)
        cast_to_new_shape = lambda tensor: tensor.view(size=(number_samples, number_datapoints) + tensor.shape[1:])
        output = {key: cast_to_new_shape(val)
                  if is_tensor(val) else map_iterable(cast_to_new_shape, val) if contains_tensors(val) else val
                  for key, val in reshaped_output.items()}
        return output

    def _get_parameters_from_input_values(self, input_values):
        """
        Method.

        Args:

        Returns:
        """
        if input_values:
            number_samples, _ = get_number_samples_and_datapoints(input_values)
        else:
            number_samples = 1
        deterministic_parents_values = {parent: parent._get_sample(number_samples, input_values=input_values)[parent] for parent in self.parents
                                        if isinstance(parent, RootVariable) or parent._type == "Deterministic node"}
        parents_input_values = {parent: parent_input for parent, parent_input in input_values.items() if parent in self.parents}
        parents_values = {**parents_input_values, **deterministic_parents_values}
        parameters_dict = self._apply_link(parents_values)
        return parameters_dict

    def _get_its_own_value_from_input(self, input_values, reevaluate):
        """
        Method.

        Args:

        Returns:
        """
        if self in input_values:
            value = input_values[self]
        elif self._type == "deterministic node":
            value = self._get_sample(1, input_values=input_values)[self]
        else:
            value = self.value
        return value

    def calculate_log_probability(self, input_values, reevaluate=True, for_gradient=False,
                                  include_parents=True, normalized=True):
        """
        Method. It returns the log probability of the values given the model. This value is always 0 since the probability
        of a deterministic variable having its value is always 1.

        Args:
            values: Dictionary(brancher.Variable: chainer.Variable). A dictionary having the brancher.variables of the
            model as keys and chainer.Variables as values. This dictionary has to provide values for all variables of
            the model except for the deterministic variables.

            reevaluate: Bool. If false it returns the output of the latest call. It avoid unnecessary computations when
            multiple children variables ask for the log probability of the same paternt variable.

        Returns:
            torch.Tensor. The log probability of the input values given the model.

        """
        if self._evaluated and not reevaluate:
            return 0.
        value = self._get_its_own_value_from_input(input_values, reevaluate)
        self._evaluated = True
        parameters_dict = self._get_parameters_from_input_values(input_values)
        log_probability = self.distribution.calculate_log_probability(value, **parameters_dict)
        parents_log_probability = sum([parent.calculate_log_probability(input_values, reevaluate, for_gradient,
                                                                        normalized=normalized)
                                       for parent in self.parents])
        if self.is_observed:
            log_probability = log_probability.sum(dim=1, keepdim=True)
        if is_tensor(log_probability) and is_tensor(parents_log_probability):
            log_probability, parents_log_probability = partial_broadcast(log_probability, parents_log_probability)
        if include_parents:
            return log_probability + parents_log_probability
        else:
            return log_probability

    def _get_statistic(self, query, input_values):
        parameters_dict = self._get_parameters_from_input_values(input_values)
        statistic = query(self.distribution, parameters_dict)
        return statistic

    def _get_sample(self, number_samples=1, resample=True, observed=False, input_values={}, differentiable=True):
        """
        Method. Used internally. It returns samples from the random variable and all its parents.

        Args:
            number_samples: . A dictionary having the brancher.variables of the
            model as keys and chainer.Variables as values. This dictionary has to provide values for all variables of
            the model except for the deterministic variables.

            resample: Bool. If false it returns the previously sampled values. It avoids that the parents of a variable
            are sampled multiple times.

            observed: Bool. It specifies if the sample should be formatted as observed data or Bayesian parameter.

            input_values: Dictionary(Variable: torch.Tensor).  dictionary of values of the parents. It is used for
            conditioning the sample on the (inputed) values of some of the parents.

        Returns:
            Dictionary(Variable: torch.Tensor). A dictionary of samples from the variable and all its parents.

        """
        if self.samples is not None and not resample:
            return {self: self.samples[-1]}
        if not observed:
            if self in input_values:
                return {self: input_values[self]}
            else:
                var_to_sample = self
        else:
            if self.has_observed_value:
                return {self: self._observed_value}
            elif self.has_random_dataset:
                var_to_sample = self.dataset
            else:
                var_to_sample = self
        parents_samples_dict = join_dicts_list([parent._get_sample(number_samples, resample, observed,
                                                                   input_values, differentiable=differentiable)
                                                for parent in var_to_sample.parents])
        input_dict = {parent: parents_samples_dict[parent] for parent in var_to_sample.parents}
        parameters_dict = var_to_sample._apply_link(input_dict)
        sample = var_to_sample.distribution.get_sample(**parameters_dict)
        self.samples = [sample] #TODO: to fix
        output_sample = {**parents_samples_dict, self: sample}
        return output_sample

    def observe(self, data):
        """
        Method. It assigns an observed value to a RandomVariable.

        Args:
            data: torch.Tensor, numeric, or np.ndarray. Input observed data.

        Returns:
            None

        """
        data = pandas_frame2value(data, self.name)
        if isinstance(data, RandomVariable):
            self.dataset = data
            self.has_random_dataset = True
        else:
            self._observed_value = coerce_to_dtype(data, is_observed=True)
            self.has_observed_value = True
        self._observed = True

    def unobserve(self):
        """
        Method. It marks a variable as not observed and it drops the dataset.

        Args:
            None

        Returns:
            None
        """
        self._observed = False
        self.has_observed_value = False
        self.has_random_dataset = False
        self._observed_value = None
        self.dataset = None

    def reset(self, recursive=True):
        """
        Method. It resets the evaluated flag of the variable and all its parents. Used after computing the
        log probability of a variable.
        """
        self.samples = None
        self._evaluated = False
        if recursive:
            for var in self.ancestors:
                var.samples = None
                var._evaluated = False

    def _flatten(self):
        variables = list(self.ancestors) + [self]
        return sorted(variables, key=lambda v: v.name)


class ProbabilisticModel(BrancherClass):
    """
    ProbabilisticModels are collections of Brancher variables.

    Parameters
    ----------
    variables: List(brancher.Variable). A list of random and deterministic variables.
    """
    def __init__(self, variables):
        self._input_variables = self._validate_variables(variables)
        self._set_summary()
        self.variables = self.flatten()
        self.posterior_model = None
        self.posterior_sampler = None
        self.observed_submodel = None
        self.is_transformed = False
        self.diagnostics = {}
        if not all([var.is_observed for var in self._input_variables]):
            self.update_observed_submodel()
        else:
            self.observed_submodel = self

    def __str__(self):
        """
        Method.

        Args: None

        Returns: String
        """
        return self.model_summary.__str__()

    @staticmethod
    def _validate_variables(variables):
        for var in variables:
            if not isinstance(var, (RootVariable, RandomVariable, ProbabilisticModel)):
                raise ValueError("Invalid input type: {}".format(type(var)))
        return variables

    def _set_summary(self):
        feature_list = ["Distribution", "Parents", "Observed"]
        var_list = self.flatten()
        var_names = [var.name for var in var_list]
        summary_data = [[var._type, var.parents, var.is_observed]
                         for var in var_list]
        self._model_summary = reformat_model_summary(summary_data, var_names, feature_list)

    @property
    def model_summary(self):
        self._set_summary()
        return self._model_summary

    @property
    def is_observed(self):
        return all([var.is_observed for var in self._flatten()])

    def observe(self, data):
        if isinstance(data, pd.DataFrame):
            data = {var_name: pandas_frame2value(data, index=var_name) for var_name in data}
        if isinstance(data, dict):
            if all([isinstance(k, Variable) for k in data.keys()]):
                data_dict = data
            if all([isinstance(k, str) for k in data.keys()]):
                data_dict = {self.get_variable(name): value for name, value in data.items()}
        else:
            raise ValueError("The input data should be either a dictionary of values or a pandas dataframe")
        for var in data_dict:
            if isinstance(var, RandomVariable):
                var.observe(data_dict[var])

    def update_observed_submodel(self):
        """
        Method. Extract the sub-model of observed variables.
        """
        flattened_model = self._flatten()
        observed_variables = [var for var in flattened_model if var.is_observed]
        self.observed_submodel = ProbabilisticModel(observed_variables)

    def set_posterior_model(self, model, sampler=None): #TODO: Clean up code duplication
        self.posterior_model = PosteriorModel(posterior_model=model, joint_model=self)
        if sampler:
            if isinstance(sampler, ProbabilisticModel):
                self.posterior_sampler = PosteriorModel(sampler, joint_model=self)
            elif isinstance(sampler, Variable):
                self.posterior_sampler = PosteriorModel(ProbabilisticModel([sampler]), joint_model=self)
            elif isinstance(sampler, Iterable) and all([isinstance(subsampler, (ProbabilisticModel, Variable))
                                                        for subsampler in sampler]):
                self.posterior_sampler = [PosteriorModel(ProbabilisticModel([var]), joint_model=self)
                                          if isinstance(var, Variable) else PosteriorModel(var, joint_model=self)
                                          for var in sampler]
            else:
                raise ValueError("The sampler should be ither a probabilistic model, a brancher variable or an iterable of variables and/or models")

    def calculate_log_probability(self, rv_values, for_gradient=False, normalized=True):
        """
        Summary
        """
        log_probability = sum([var.calculate_log_probability(rv_values, reevaluate=False,
                                                             for_gradient=for_gradient,
                                                             normalized=normalized)
                               for var in self._input_variables])
        self.reset()
        return log_probability

    def _get_statistic(self, query, input_values):
        return {var: var._get_statistic(query, input_values) for var in self.variables}

    def _get_sample(self, number_samples, observed=False, input_values={}, differentiable=True):
        """
        Summary
        """
        joint_sample = join_dicts_list([var._get_sample(number_samples=number_samples, resample=False,
                                                        observed=observed, input_values=input_values,
                                                        differentiable=differentiable)
                                        for var in self._input_variables])
        joint_sample.update(input_values)
        self.reset()
        return joint_sample

    def _get_entropy(self, input_values={}, for_gradient=True):
        if not self.is_transformed:
            entropy_array = {var: var._get_entropy(input_values) for var in self.variables}
            return sum([sum_from_dim(var_ent, 2) for var_ent in entropy_array.values()])
        else:
            return -self.calculate_log_probability(input_values, for_gradient=for_gradient)

    def get_sample(self, number_samples, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                                            number_samples=number_samples)
        raw_sample = self._get_sample(number_samples, observed=False, input_values=reformatted_input_values,
                                      differentiable=False)
        sample = reformat_sample_to_pandas(raw_sample)
        return sample

    def get_mean(self, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=1)
        raw_mean = self._get_mean(reformatted_input_values)
        mean = reformat_sample_to_pandas(raw_mean, number_samples=1)
        return mean

    def get_variance(self, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=1)
        raw_variance = self._get_variance(reformatted_input_values)
        variance = reformat_sample_to_pandas(raw_variance, number_samples=1)
        return variance

    def get_entropy(self, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                          number_samples=1)
        raw_ent = self._get_entropy(reformatted_input_values)
        ent = reformat_sample_to_pandas(raw_ent, number_samples=1)
        return ent

    def check_posterior_model(self):
        """
        Summary
        """
        if not self.posterior_model:
            raise AttributeError("The posterior model has not been initialized.")

    def get_posterior_mean(self, query, input_values):
        return self.posterior_model.get_mean(input_values)

    def get_posterior_variance(self, query, input_values):
        return self.posterior_model.get_variance(input_values)

    def get_posterior_entropy(self, query, input_values):
        return self.posterior_model.get_entropy(input_values)

    def _get_posterior_sample(self, number_samples, input_values={}, differentiable=True):
        """
        Summary
        """
        self.check_posterior_model()
        posterior_sample = self.posterior_model._get_posterior_sample(number_samples=number_samples,
                                                                      input_values=input_values,
                                                                      differentiable=differentiable)
        sample = self._get_sample(number_samples, input_values=posterior_sample, differentiable=differentiable)
        return sample

    def get_posterior_sample(self, number_samples, input_values={}):
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                                            number_samples=number_samples)
        raw_sample = self._get_posterior_sample(number_samples, input_values=reformatted_input_values)
        sample = reformat_sample_to_pandas(raw_sample)
        return sample

    def get_p_log_probabilities_from_q_samples(self, q_samples, q_model, empirical_samples={},
                                               for_gradient=False, normalized=True):
        p_samples = reassign_samples(q_samples, source_model=q_model, target_model=self)
        p_samples.update(empirical_samples)
        p_log_prob = self.calculate_log_probability(p_samples, for_gradient=for_gradient, normalized=normalized)
        return p_log_prob

    def get_importance_weights(self, q_samples, q_model, empirical_samples={},
                               for_gradient=False, give_normalization=False):
        if not empirical_samples:
            empirical_samples = self.observed_submodel._get_sample(1, observed=True, differentiable=False)
        q_log_prob = q_model.calculate_log_probability(q_samples,
                                                       for_gradient=for_gradient, normalized=False)
        p_log_prob = self.get_p_log_probabilities_from_q_samples(q_samples=q_samples,
                                                                 q_model=q_model,
                                                                 empirical_samples=empirical_samples,
                                                                 for_gradient=for_gradient,
                                                                 normalized=False)
        log_weights = (p_log_prob - q_log_prob).detach().numpy()
        alpha = np.max(log_weights)
        norm_log_weights = log_weights - alpha
        weights = np.exp(norm_log_weights)
        norm = np.sum(weights)
        weights /= norm
        if not give_normalization:
            return weights
        else:
            return weights, np.log(norm) + alpha #TODO: work in progress

    def estimate_log_model_evidence(self, number_samples, method="ELBO", input_values={},
                                    for_gradient=False, posterior_model=(), gradient_estimator=None):
        if not posterior_model:
            self.check_posterior_model()
            posterior_model = self.posterior_model
        if method is "ELBO":
            empirical_samples = self.observed_submodel._get_sample(1, observed=True, differentiable=False) #TODO Important!!: You need to correct for subsampling
            if for_gradient:
                function = lambda samples: self.get_p_log_probabilities_from_q_samples(q_samples=samples,
                                                                                       empirical_samples=empirical_samples,
                                                                                       for_gradient=for_gradient,
                                                                                       q_model=posterior_model) + posterior_model._get_entropy(samples,
                                                                                                                                               for_gradient=for_gradient)
                estimator = gradient_estimator(function, posterior_model, empirical_samples)
                log_model_evidence = estimator(number_samples)
            else:
                posterior_samples = posterior_model._get_sample(number_samples=number_samples,
                                                                observed=False, input_values=input_values,
                                                                differentiable=False)
                joint_log_prob = self.get_p_log_probabilities_from_q_samples(q_samples=posterior_samples,
                                                                             empirical_samples=empirical_samples,
                                                                             for_gradient=for_gradient,
                                                                             q_model=posterior_model)
                posterior_entropy = posterior_model._get_entropy(posterior_samples, for_gradient=for_gradient)
                log_model_evidence = torch.mean(joint_log_prob + posterior_entropy)
            return log_model_evidence
        else:
            raise NotImplementedError("The requested estimation method is currently not implemented.")

    def reset(self):
        """
        Summary
        """
        for var in self.flatten():
            var.reset(recursive=False)

    def _flatten(self):
        variables = list(join_sets_list([var.ancestors.union({var}) for var in self._input_variables]))
        return sorted(variables, key=lambda v: v.name)


class PosteriorModel(ProbabilisticModel):
    """
    Summary

    Parameters
    ----------
    variables : tuple of brancher variables
        Summary
    """
    def __init__(self, posterior_model, joint_model):
        super().__init__(posterior_model.variables)
        self.posterior_model = None
        self.model_mapping = get_model_mapping(self, joint_model)

        self._is_trained = False

    def posterior_sample2joint_sample(self, posterior_sample):
        return reassign_samples(posterior_sample, self.model_mapping)

    def _get_posterior_sample(self, number_samples, observed=False, input_values={}, differentiable=True):
        sample = self.posterior_sample2joint_sample(self._get_sample(number_samples, observed, input_values,
                                                                     differentiable=differentiable))
        sample.update(input_values)
        return sample


def var2link(var):
    if isinstance(var, Variable):
        vars = {var}
        fn = lambda values: values[var]
    elif isinstance(var, (numbers.Number, np.ndarray, torch.Tensor)):
        vars = set()
        fn = lambda values: var
    elif isinstance(var, (tuple, list)) and all([isinstance(v, (Variable, PartialLink)) for v in var]):
        vars = join_sets_list([{v} if isinstance(v, Variable) else v.vars for v in var])
        fn = lambda values: tuple([values[v] if isinstance(v, Variable) else v.fn(values) for v in var])
    else:
        return var
    return PartialLink(vars=vars, fn=fn, links=set(), string=str(var))


class Ensemble(BrancherClass):

    def __init__(self, model_list, weights=None):
        #TODO: assert that all variables have the same name
        self.num_models = len(model_list)
        self.model_list = model_list
        if weights is None:
            self.weights = [1./self.num_models]*self.num_models
        else:
            self.weights = np.array(weights)

    def _get_sample(self, number_samples, observed=False, input_values={}, differentiable=True):
        num_samples_list = np.random.multinomial(number_samples, self.weights)
        samples_list = [model._get_sample(n, differentiable=differentiable)
                        for n, model in zip(num_samples_list, self.model_list)]
        named_sample_list = [{var.name: value for var, value in sample.items()} for sample in samples_list]
        named_sample = concatenate_samples(named_sample_list)
        sample = {self.model_list[0].get_variable(name): value for name, value in named_sample.items()}
        return sample

    # def calculate_log_probability(self, rv_values, for_gradient=False, normalized=True):
    #     """
    #     Summary
    #     """
    #     log_probabilities = [model.calculate_log_probability(rv_values, reevaluate=False,
    #                                                          for_gradient=for_gradient,
    #                                                          normalized=normalized)
    #                          for model in self.model_list]
    #     alpha = np.max(log_probabilities)


    def _flatten(self):
        return [model.flatten() for model in self.model_list]

    def get_sample(self, number_samples, input_values={}): #TODO: code duplication here
        reformatted_input_values = reformat_sampler_input(pandas_frame2dict(input_values),
                                                                            number_samples=number_samples)
        raw_sample = self._get_sample(number_samples, observed=False, input_values=reformatted_input_values,
                                      differentiable=False)
        sample = reformat_sample_to_pandas(raw_sample)
        return sample

    def _get_statistic(self, query, input_values):
        raise NotImplemented

    def _get_mean(self, input_values):
        raise NotImplemented

    def _get_variance(self, input_values):
        raise NotImplemented


class PartialLink(BrancherClass):

    def __init__(self, vars, fn, links, string=""):
        self.vars = vars
        self.fn = fn
        self.links = links
        self.string = string

    def __str__(self):
        """
        Method.

        Args: None

        Returns: String
        """
        return self.string

    def _apply_operator(self, other, op):
        symbols = {operator.add: "+", operator.sub: "-", operator.mul: "*", operator.truediv: "/", operator.pow: "**"}
        get_symbol = lambda op: symbols[op] if op in symbols.keys() else "?"
        other = var2link(other)
        return PartialLink(vars=self.vars.union(other.vars),
                           fn=lambda values: op(self.fn(values), other.fn(values)),
                           links=self.links.union(other.links),
                           string="(" + str(self) + get_symbol(op) + str(other) + ")")

    def __neg__(self):
        return -1*self

    def __add__(self, other):
        return self._apply_operator(other, operator.add)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self._apply_operator(other, operator.sub)

    def __rsub__(self, other):
        return -1*self.__sub__(other)

    def __mul__(self, other):
        return self._apply_operator(other, operator.mul)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._apply_operator(other, operator.truediv)

    def __rtruediv__(self, other):
        return self.__truediv__(other)**(-1)

    def __pow__(self, other):
        return self._apply_operator(other, operator.pow)

    def __rpow__(self, other):
        raise NotImplementedError

    def __getitem__(self, key):
        if isinstance(key, collections.Iterable) and all([isinstance(k, int) for k in key]):
            variable_slice = (slice(None, None, None), *key)
        elif isinstance(key, int):
            variable_slice = (slice(None, None, None), key)
        elif isinstance(key, collections.Hashable):
            variable_slice = key
        else:
            raise ValueError("The input to __getitem__ is neither numeric nor a hashabble key")

        vars = self.vars
        fn = lambda values: self.fn(values)[variable_slice] if is_tensor(self.fn(values)) \
            else self.fn(values)[key]
        links = set()
        return PartialLink(vars=vars,
                           fn=fn,
                           links=self.links)

    def shape(self):
        vars = self.vars
        fn = lambda values: self.fn(values).shape
        links = set()
        return PartialLink(vars=vars,
                           fn=fn,
                           links=self.links)

    def _flatten(self):
        return flatten_list([var._flatten() for var in self.vars]) + [self]

    def _get_statistic(self, query, input_values):
        raise NotImplemented

    def _get_mean(self, input_values):
        raise NotImplemented

    def _get_variance(self, input_values):
        raise NotImplemented


"""
Config
---------
Set device to 'cpu' or 'cuda:index'
Default: cuda:0 if available, otherwise cpu

"""
import torch

default_device = 'cpu'

def set_device(device_):
    global device
    device_ = device_.lower()
    if isinstance(device_, str):
        if 'cuda' in device_:
            assert torch.cuda.is_available(), "Cuda requested but not available"
            device = torch.device(device_)
        elif device_ == 'cpu':
            device = torch.device(device_)
        elif device_ == 'gpu':
            assert torch.cuda.is_available(), "Cuda requested but not available"
            device = torch.device('cuda:0')
        else:
            raise ValueError("Device is not recongnized")
    elif isinstance(device_, int):
        device = torch.device(device_)

if 'device' not in locals():
    set_device(default_device)




import copy

import numpy as np

from brancher.variables import RandomVariable, ProbabilisticModel
from brancher.utilities import concatenate_samples, reject_samples
from brancher.utilities import is_tensor


def truncate_model(model, truncation_rule, model_statistics):

    def truncated_calculate_log_probability(rv_values, for_gradient=False, normalized=True):
        unnormalized_log_probability = model.calculate_log_probability(rv_values, normalized=normalized,
                                                                       for_gradient=for_gradient)
        if not normalized:
            return unnormalized_log_probability
        else:
            if for_gradient:
                nondiff_values = {var: value.detach() if is_tensor(value) else value for var, value in rv_values.items()}
                normalization = -model.calculate_log_probability(nondiff_values,
                                                                 for_gradient=False, normalized=True).mean()
                return unnormalized_log_probability + normalization
            else:
                number_samples = list(rv_values.values())[0].shape[0]
                acceptance_ratio = get_acceptance_probability(number_samples=number_samples)
                return unnormalized_log_probability - np.log(acceptance_ratio)

    def truncated_get_sample(number_samples, max_itr=1, **kwargs):
        batch_size = number_samples
        current_number_samples = 0
        sample_list = []
        itr = 0
        while (current_number_samples < number_samples and itr < max_itr) or current_number_samples < 1:
            remaining_samples, n, p = reject_samples(model._get_sample(batch_size, **kwargs),
                                                     model_statistics=model_statistics,
                                                     truncation_rule=truncation_rule)
            if remaining_samples:
                remaining_samples = {var: value[:number_samples - current_number_samples, :]
                                     for var, value in remaining_samples.items()}
                current_number_samples += np.min([n, number_samples - current_number_samples])
                sample_list.append(remaining_samples)
            itr += 1
        return concatenate_samples(sample_list)

    def get_acceptance_probability(samples=None, number_samples=None):
        if not samples:
            samples = model._get_sample(number_samples)
        _, _, p = reject_samples(samples,
                                 model_statistics=model_statistics,
                                 truncation_rule=truncation_rule)
        return p

    truncated_model = copy.copy(model)

    if isinstance(model, ProbabilisticModel):
        truncated_model.is_transformed = True
        truncated_model._get_sample = truncated_get_sample
        truncated_model.calculate_log_probability = truncated_calculate_log_probability
        truncated_model.get_acceptance_probability = get_acceptance_probability
        for var in model.variables:
            var.distribution.has_analytic_entropy = False
            var.distribution.has_analytic_mean = False
            var.distribution.has_analytic_var = False

    elif isinstance(model, RandomVariable):
        pass #TODO: Work in progress
    else:
        raise ValueError("Only probabilistic models and random variables can be truncated")
    return truncated_model



import pandas as pd
import numpy as np
from brancher.utilities import is_tensor
from brancher.utilities import map_iterable
from collections.abc import Iterable


def pandas_dict2list(dic):
    indices, values = zip(*dic.items())
    sorted_indices = np.argsort(indices)
    return np.array(values)[sorted_indices]


def pandas_frame2dict(dataframe):
    if isinstance(dataframe, pd.core.frame.DataFrame):
        return {key: pandas_dict2list(val) for key, val in dataframe.to_dict().items()}
    elif isinstance(dataframe, dict):
        return dataframe
    else:
        raise ValueError("The input should be either a dictionary or a Pandas dataframe")


def pandas_frame2value(dataframe, index):
    if isinstance(dataframe, pd.core.frame.DataFrame):
        values = np.array([np.ndarray.tolist(x) if isinstance(x, np.ndarray) else x
                           for x in dataframe[index].values]) #TODO: Ugly, this should be improved
        return values
    else:
        return dataframe


def reformat_value(value, index):
    if is_tensor(value):
        if np.prod(value[index, :].shape) == 1:
            return float(value[index, :].cpu().detach().numpy())
        else: #elif value.shape[1] == 1:
            return value[index, :].cpu().detach().numpy()#[0, :]
        #else:
        #    return value.cpu().detach().numpy()
    elif isinstance(value, Iterable):
        return map_iterable(lambda x: reformat_value(x, index), value)
    else:
        return value


def reformat_sample_to_pandas(sample):
    number_samples = list(sample.values())[0].shape[0]
    data = [[reformat_value(value, index)
             for index in range(number_samples)]
            for variable, value in sample.items()]
    index = [key.name for key in sample.keys()]
    column = range(number_samples)
    frame = pd.DataFrame(data, index=index, columns=column).transpose()
    return frame


def reformat_model_summary(summary_data, var_names, feature_list):
    return pd.DataFrame(summary_data, index=var_names, columns=feature_list).transpose()



"""
Utilities
---------
Module description
"""
import sys
from functools import reduce
from collections import abc
from collections.abc import Iterable

import numpy as np
import torch
import pandas as pd

from brancher.config import device


def is_tensor(data):
    return torch.is_tensor(data)


def contains_tensors(data):
    if isinstance(data, dict):
        return all([is_tensor(d) for d in data.values()])
    if isinstance(data, Iterable):
        return all([is_tensor(d) for d in data])
    else:
        return False


def is_discrete(data):
    return type(data) in [list, set, tuple, dict, str]


def to_tuple(obj):
    if isinstance(obj, Iterable) and not isinstance(obj, torch.Tensor):
        return tuple(obj)
    else:
        return tuple([obj])


def to_tensor(arr):
    if isinstance(arr, torch.Tensor):
        return arr
    elif isinstance(arr, np.ndarray):
        return torch.Tensor(arr)
    else:
        raise ValueError("The input should be either a torch.Tensor or a np.array")


def map_iterable(func, itr, recursive=False):
    def f(x):
        if not recursive:
            return func(x)
        else:
            return map_iterable(func, x, recursive=True)
    if is_tensor(itr) or not isinstance(itr, Iterable):
        return func(itr)
    if isinstance(itr, dict):
        return {key: f(val) for key, val in itr.items()}
    out = [*map(f, itr)]
    if isinstance(itr, list):
        return out
    elif isinstance(itr, tuple):
        return tuple(out)

def zip_dict(first_dict, second_dict):
    keys = set(first_dict.keys()).intersection(set(second_dict.keys()))
    return {k: to_tuple(first_dict[k]) + to_tuple(second_dict[k]) for k in keys}


def zip_dict_list(dict_list):
    if len(dict_list) == 0:
        return {}
    if len(dict_list) == 1:
        return dict_list[0]
    else:
        zipped_dict = zip_dict(dict_list[-1], dict_list[-2])
        new_dict_list = dict_list[:-2]
        new_dict_list.append(zipped_dict)
        return zip_dict_list(new_dict_list)


def split_dict(dic, condition):
    dict_1 = {}
    dict_2 = {}
    for key, val in dic.items():
        if condition(key, val):
            dict_1.update({key: val})
        else:
            dict_2.update({key: val})
    return dict_1, dict_2


def flatten_list(lst):
    flat_list = [item for sublist in lst for item in sublist]
    return flat_list


def flatten_set(st):
    flat_set = set([item for subset in st for item in subset])
    return flat_set


def join_dicts_list(dicts_list):
    if dicts_list:
        return reduce(lambda d1, d2: {**d1, **d2}, dicts_list)
    else:
        return {}


def join_sets_list(sets_list):
    if sets_list:
        return reduce(lambda d1, d2: d1.union(d2), sets_list)
    else:
        return set()


def sum_from_dim(tensor, dim_index):
    if is_tensor(tensor):
        data_dim = len(tensor.shape)
        for dim in reversed(range(dim_index, data_dim)):
            tensor = tensor.sum(dim=dim)
        return tensor
    else:
        return np.sum(tensor, axis=tuple(range(1, len(tensor.shape) - 1)), keepdims=False)[:, 0]

def sum_data_dimensions(var):
    return sum_from_dim(var, dim_index=2)


def partial_broadcast(*args):
    assert all([is_tensor(ar) for ar in args]), 'at least 1 object is not torch tensor'
    shapes0, shapes1 = zip(*[(x.shape[0], x.shape[1]) for x in args])
    s0, s1 = np.max(shapes0), np.max(shapes1)
    return [x.expand((s0, s1) + x.shape[2:]) for x in args]


def tile_batch_dimensions(tensor, number_samples, number_datapoints):
    return tensor.expand((number_samples, number_datapoints) + tensor.shape[2:])


def broadcast_and_squeeze(*args):
    assert all([is_tensor(ar) for ar in args]), 'at least 1 object is not torch tensor'
    if all([np.prod(val.shape[2:]) == 1 for val in args]):
        args = [val.contiguous().view(size=val.shape[:2] + tuple([1, 1])) for val in args]
    uniformed_values = uniform_shapes(*args)
    broadcasted_values = torch.broadcast_tensors(*uniformed_values)
    return broadcasted_values


def broadcast_and_squeeze_mixed(tpl, dic):
    tpl_len = len(tpl)
    dict_keys, dict_values = zip(*dic.items())
    broadcasted_values = broadcast_and_squeeze(*(tpl + dict_values))
    if tpl_len > 0:
        return broadcasted_values[:tpl_len], {k: v for k, v in zip(dict_keys, broadcasted_values[tpl_len:])}
    else:
        return {k: v for k, v in zip(dict_keys, broadcasted_values[tpl_len:])}


def get_items(itr, recursive=False):
    if is_tensor(itr) or not isinstance(itr, Iterable):
        return iter

    def f(x):
        if recursive:
            return get_items(x, recursive=True)
        else:
            return x

    if isinstance(itr, dict):
        items = f(itr.items())
        itr.items()
    else:
        return f(itr)


def reshape_parent_value(value, number_samples, number_datapoints):
    newshape = tuple([number_samples * number_datapoints]) + value.shape[2:]
    return value.contiguous().view(size=newshape)


def broadcast_and_reshape_parent_value(value, number_samples, number_datapoints):
    return reshape_parent_value(tile_batch_dimensions(value, number_samples, number_datapoints),
                                number_samples, number_datapoints)


def get_number_samples_and_datapoints(parent_values):
    n_list = []
    m_list = []
    for value in parent_values.values():
        if is_tensor(value):
            n_list.append(value.shape[0])
            m_list.append(value.shape[1])
        elif contains_tensors(value):
            if isinstance(value, dict):
                itr = value.values()
            else:
                itr = value
            for tensor in itr:
                n_list.append(tensor.shape[0])
                m_list.append(tensor.shape[1])
    if not n_list and not m_list:
        return None, None
    else:
        return max(n_list), max(m_list)


def get_diagonal(tensor):
    assert torch.is_tensor(tensor), 'object is not torch tensor'
    assert tensor.ndimension() == 4, 'ndim should be equal 4'
    dim1, dim2, dim_matrix, _ = tensor.shape
    diag_ind = list(range(dim_matrix))
    expanded_diag_ind = dim1*dim2*diag_ind
    axis12_ind = [a for a in range(dim1*dim2) for _ in range(dim_matrix)]
    reshaped_tensor = tensor.contiguous().view(size=(dim1*dim2, dim_matrix, dim_matrix))
    ind = (np.array(axis12_ind), np.array(expanded_diag_ind), np.array(expanded_diag_ind))
    subdiagonal = reshaped_tensor[ind]
    return subdiagonal.view(size=(dim1, dim2, dim_matrix))


def coerce_to_dtype(data, is_observed=False):
    """Summary"""
    def reformat_tensor(result):
        if is_observed:
            result = torch.unsqueeze(result, dim=0)
            result_shape = result.shape
            if len(result_shape) == 2:
                result = result.contiguous().view(size=result_shape + tuple([1, 1]))
            elif len(result_shape) == 3:
                result = result.contiguous().view(size=result_shape + tuple([1]))
            #if len(result_shape) == 2:
            #   result = result.contiguous().view(size=result_shape + tuple([1]))
        else:
            result = torch.unsqueeze(torch.unsqueeze(result, dim=0), dim=1)
        return result

    dtype = type(data) ##TODO: do we need any additional shape checking?
    if dtype is torch.Tensor: # to tensor
        result = data.float()
    elif dtype is np.ndarray: # to tensor
        result = torch.tensor(data).float()
    elif dtype is pd.DataFrame:
        result = torch.tensor(data.values).float()
    elif dtype in [float, int] or dtype.__base__ in [np.floating, np.signedinteger]: # to tensor
        result = torch.tensor(data * np.ones(shape=(1, 1))).float()
    elif dtype in [list, set, tuple, dict, str]: # to discrete
        return data
    else:
        raise TypeError("Invalid input dtype {} - expected float, integer, np.ndarray, or torch var.".format(dtype))

    result = reformat_tensor(result)
    return result.to(device)


def tile_parameter(tensor, number_samples):
    assert is_tensor(tensor), 'object is not torch tensor'
    value_shape = tensor.shape
    if value_shape[0] == number_samples:
        return tensor
    elif value_shape[0] == 1:
        reps = tuple([number_samples] + [1] * len(value_shape[1:]))
        return tensor.repeat(repeats=reps)
    else:
        raise ValueError("The parameter cannot be broadcasted to the required number of samples")


def reformat_sampler_input(sample_input, number_samples):
    return {var: tile_parameter(coerce_to_dtype(value, is_observed=var.is_observed), number_samples=number_samples)
            for var, value in sample_input.items()}


def uniform_shapes(*args):
    assert all([is_tensor(ar) for ar in args]), 'at least 1 object is not torch tensor'
    shapes = [ar.shape for ar in args]
    max_len = np.max([len(s) for s in shapes])
    return [torch.unsqueeze(ar, dim=len(ar.shape)) if (len(ar.shape) == max_len-1) else ar
            for ar in args]


def get_model_mapping(source_model, target_model):
    model_mapping = {}
    if isinstance(target_model, dict):
        target_variables = list(target_model.keys())
    else:
        target_variables = target_model._flatten()
    for p_var in target_variables:
        try:
            model_mapping.update({source_model.get_variable(p_var.name): p_var})
        except KeyError:
            pass
    return model_mapping


def reassign_samples(samples, model_mapping=(), source_model=(), target_model=()):
    out_sample = {}
    if model_mapping:
        pass
    elif source_model and target_model:
        model_mapping = get_model_mapping(source_model, target_model)
    else:
        raise ValueError("Either a model mapping or both source and target models have to be provided as input")
    for key, value in samples.items():
        try:
            out_sample.update({model_mapping[key]: value})
        except KeyError:
            pass
    return out_sample


def reject_samples(samples, model_statistics, truncation_rule):
    decision_variable = model_statistics(samples)
    sample_indices = [index for index, value in enumerate(decision_variable) if truncation_rule(value)]
    num_accepted_samples = len(sample_indices)
    if num_accepted_samples == 0:
        return None, 0, 0.001 #TODO: Improve
    else:
        remaining_samples = {var: value[sample_indices, :] for var, value in samples.items()}

        acceptance_probability = num_accepted_samples/float(decision_variable.shape[0])
        return remaining_samples, num_accepted_samples, acceptance_probability


def concatenate_samples(samples_list):
    ''' replaced with torch.cat'''
    if len(samples_list) == 1:
        return samples_list[0]
    else:
        #num_samples = len(samples_list)
        paired_list = zip_dict_list(samples_list)
        samples = {var: torch.cat(tensor_tuple, dim=0)#.contiguous().view(tuple([num_samples]) + tuple(tensor_tuple[0].shape[1:]))
                   for var, tensor_tuple in paired_list.items()}
        return samples


def tensor_range(tensor):
    return set(np.ndarray.tolist(tensor.detach().numpy().flatten()))


def batch_meshgrid(tensor1, tensor2):
    tensor1_shape = tensor1.shape
    tensor2_shape = tensor2.shape
    new_shape = [tensor1_shape[0], tensor1_shape[1], tensor2_shape[1]]

    assert (len(tensor1_shape) == 2 and len(tensor2_shape) == 2), "You can use batch_meshgrid only on 2D tensor (The first dimension is the batch dimension)"

    tensor1 = tensor1.unsqueeze(dim=2).expand(*new_shape)
    tensor2 = tensor2.unsqueeze(dim=1).expand(*new_shape)
    return tensor1, tensor2


def get_tensor_data(tensor):
    return tensor.cpu().detach().numpy()


def delta(x, y):
    return (x == y).float()


"""
---------
"""
from abc import ABC, abstractmethod
import numpy as np

import brancher.functions as BF
from brancher.utilities import partial_broadcast


class GeometricRange(ABC):

    @abstractmethod
    def forward_transform(self, x, dim):
        pass

    @abstractmethod
    def inverse_transform(self, x, dim):
        pass


class UnboundedRange(GeometricRange):

    def __init__(self):
        pass

    def forward_transform(self, x, dim):
        return x

    def inverse_transform(self, y, dim):
        return y


class Interval(GeometricRange):

    def __init__(self, lower_bound, upper_bound):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def forward_transform(self, x, dim):
        return self.lower_bound + (self.upper_bound-self.lower_bound)*BF.sigmoid(x)

    def inverse_transform(self, y, dim):
        z = (y - self.lower_bound) / (self.upper_bound - self.lower_bound)
        return np.log(z / (1 - z))


class RightHalfLine(GeometricRange):

    def __init__(self, lower_bound):
        self.lower_bound = lower_bound

    def forward_transform(self, x, dim):
        return self.lower_bound + BF.softplus(x)

    def inverse_transform(self, y, dim):
        return np.log(np.exp(y - self.lower_bound) - 1)


class LeftHalfLine(GeometricRange):

    def __init__(self, upper_bound):
        self.lower_bound = upper_bound

    def forward_transform(self, x, dim):
        return self.lower_bound - BF.softplus(x)

    def inverse_transform(self, y, dim):
        return np.log(np.exp(-y + self.lower_bound) - 1)


class Simplex(GeometricRange):

    def forward_transform(self, x, dim):
        latent_p = BF.softplus(x)
        normalization = BF.sum(latent_p, axis=1, keepdims=True)
        normalization = BF.broadcast_to(normalization, latent_p.shape())
        return latent_p / normalization

    def inverse_transform(self, y, dim):
        return np.log(np.exp(y) - 1)


class PositiveDefiniteMatrix(GeometricRange): #TODO: Work in progress

    def forward_transform(self, x, dim):
        return BF.matmul(x, BF.transpose(x, -2, -1))

    def inverse_transform(self, y, dim):
        chol_factor = np.linalg.cholesky(y)
        return chol_factor


# class UpperTriangularPositiveDefiniteMatrix(GeometricRange): #TODO: Work in progress
#
#     def forward_transform(self, x, dim):
#         pass
#         #matrix_shape = x.shape[-2:]
#         #return BF.matmul(x, BF.transpose(x, -2,-1))
#
#     def inverse_transform(self, y, dim):
#         pass
#         #chol_factor = np.linalg.cholesky(y)
#         #return chol_factor




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='brancher',
      version='0.3.0',
      description='A user-centered Python package for differentiable probabilistic inference',
      author='Brancher development team',
      author_email='info@brancher.org',
      license='MIT',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://brancher.org/",
      packages=find_packages(),
      install_requires=['python_version>="3.7"',
                        'numpy>=1.15.4',
                        'pandas>=0.23.4',
                        'matplotlib>=3.0.2',
                        'seaborn>=0.9.0',
                        'scipy>=1.1.0',
                        'tqdm>=4.28.1',
                        ],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      )

__author__ = 'MindCodec'



import numpy as np


from brancher.variables import RootVariable
from brancher.utilities import is_tensor

class VoronoiSet(object):

    def __init__(self, particles, cost=lambda x, y: np.sum((x - y)**2)):
        self.cost = cost
        self.particles = particles
        self.locations = None
        self.update_locations()

    def update_locations(self):
        if isinstance(self.particles, list):
            if is_tensor(self.particles[0]):
                self.locations = [part.data for part in self.particles]
            elif isinstance(self.particles[0], RootVariable):
                self.locations = [part.value[0, 0, :].data for part in self.particles]
            else:
                raise ValueError("The location of the particles should be either deterministic brancher variables, chainer variables or np.array")
        else:
            raise ValueError("The location of the particles should be inserted as a list of locations")

    def __call__(self, x, index):
        self.update_locations()
        distances = [self.cost(x, y) for y in self.locations]
        if np.argmin(distances) == index:
            return True
        else:
            return False




"""
Inference
---------
Module description
"""
from abc import ABC, abstractmethod

import numpy as np

import torch

from brancher.config import device
from brancher.utilities import map_iterable
from brancher.utilities import zip_dict_list


class GradientEstimator(ABC):

    def __init__(self, function, sampler, empirical_samples={}):
        self.function = function
        self.sampler = sampler
        self.empirical_samples = empirical_samples

    @abstractmethod
    def __call__(self, samples):
        pass


class BlackBoxEstimator(GradientEstimator):

    def __call__(self, n_samples):
        samples = self.sampler._get_sample(n_samples, differentiable=False)
        samples.update(self.empirical_samples)
        variational_loss = self.sampler.calculate_log_probability(samples)*(self.function(samples).detach())
        model_loss = self.function(samples)
        return (variational_loss + model_loss).mean()


class PathwiseDerivativeEstimator(GradientEstimator):

    def __call__(self, n_samples):
        samples = self.sampler._get_sample(n_samples, differentiable=True)
        samples.update(self.empirical_samples)
        return self.function(samples).mean()


class Taylor1Estimator(GradientEstimator):

    def __call__(self, n_samples):
        observable_samples = self.sampler._get_sample(n_samples, differentiable=False)
        observable_samples.update(self.empirical_samples)
        means = {var: var._get_mean(input_values=observable_samples) for var in self.sampler.variables if not var.is_observed}
        for key, value in observable_samples.items():
            if not key in means:
                means[key] = value
        return self.function(means).mean()



import numbers

import numpy as np
import torch.nn as nn

import brancher.distributions as distributions
import brancher.functions as BF
import brancher.geometric_ranges as geometric_ranges
from brancher.variables import var2link, Variable, RootVariable, RandomVariable, PartialLink
from brancher.utilities import join_sets_list



class LinkConstructor(nn.ModuleList):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        modules = [link
                 for partial_link in kwargs.values()
                 for link in var2link(partial_link).links]
        super().__init__(modules) #TODO: asserts that specified links are valid pytorch modules

    def __call__(self, values):
        return {k: var2link(x).fn(values) for k, x in self.kwargs.items()}


class VariableConstructor(RandomVariable):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, name, learnable, ranges, is_observed=False, **kwargs): #TODO: code duplication here
        self.name = name
        self._evaluated = False
        self._observed = is_observed
        self._observed_value = None
        self._current_value = None
        self.construct_deterministic_parents(learnable, ranges, kwargs)
        self.parents = join_sets_list([var2link(x).vars for x in kwargs.values()])
        self.ancestors = join_sets_list([self.parents] + [parent.ancestors for parent in self.parents])
        self.link = LinkConstructor(**kwargs)
        self.samples = None
        self.ranges = {}
        self.dataset = None
        self.has_random_dataset = False
        self.has_observed_value = False
        self.is_normalized = True
        self.partial_links = {name: var2link(link) for name, link in kwargs.items()}

    def construct_deterministic_parents(self, learnable, ranges, kwargs):
        for parameter_name, value in kwargs.items():
            if not isinstance(value, (Variable, PartialLink)):
                if isinstance(value, np.ndarray):
                    dim = value.shape[0]
                elif isinstance(value, numbers.Number):
                    dim = 1
                else:
                    dim = []
                deterministic_parent = RootVariable(ranges[parameter_name].inverse_transform(value, dim),
                                                    self.name + "_" + parameter_name, learnable, is_observed=self._observed)
                kwargs.update({parameter_name: ranges[parameter_name].forward_transform(deterministic_parent, dim)})


class EmpiricalVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, dataset, name, learnable=False, is_observed=False, batch_size=None, indices=None, weights=None): #TODO: Ugly logic
        self._type = "Empirical"
        input_parameters = {"dataset": dataset, "batch_size": batch_size, "indices": indices, "weights": weights}
        ranges = {par_name: geometric_ranges.UnboundedRange()
                  for par_name, par_value in input_parameters.items()
                  if par_value is not None}
        kwargs = {par_name: par_value
                  for par_name, par_value in input_parameters.items()
                  if par_value is not None}
        super().__init__(name, **kwargs, learnable=learnable, ranges=ranges, is_observed=is_observed)

        if not batch_size:
            if indices:
                batch_size = len(indices)
            else:
                raise ValueError("Either the indices or the batch size has to be given as input")

        self.batch_size = batch_size
        self.distribution = distributions.EmpiricalDistribution(batch_size=batch_size, is_observed=is_observed)


class RandomIndices(EmpiricalVariable):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, dataset_size, batch_size, name, is_observed=False):
        self._type = "Random Index"
        super().__init__(dataset=list(range(dataset_size)),
                         batch_size=batch_size, is_observed=is_observed, name=name)

    def __len__(self):
        return self.batch_size


class DeterministicVariable(VariableConstructor): #TODO: Future refactor? Should Deterministic variables and deterministic node be different? (No probably not)
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, value, name, learnable=False, is_observed=False, variable_range=geometric_ranges.UnboundedRange()):
        self._type = "Deterministic node"
        ranges = {"value": variable_range}
        super().__init__(name, value=value, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.DeterministicDistribution()

    @property
    def value(self):
        return self._get_sample(1)[self]


class NormalVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, loc, scale, name, learnable=False, is_observed=False):
        self._type = "Normal"
        ranges = {"loc": geometric_ranges.UnboundedRange(),
                  "scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.NormalDistribution()


class StudentTVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, df, loc, scale, name, learnable=False, is_observed=False):
        self._type = "StudentT"
        ranges = {"df": geometric_ranges.UnboundedRange(),
                  "loc": geometric_ranges.UnboundedRange(),
                  "scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, df=df, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.StudentTDistribution()


class UniformVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, low, high, name, learnable=False, is_observed=False):
        self._type = "Uniform"
        ranges = {"low": geometric_ranges.UnboundedRange(),
                  "high": geometric_ranges.UnboundedRange()}
        super().__init__(name, low=low, high=high, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.UniformDistribution()


class CauchyVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, loc, scale, name, learnable=False, is_observed=False):
        self._type = "Cauchy"
        ranges = {"loc": geometric_ranges.UnboundedRange(),
                  "scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.CauchyDistribution()


class HalfCauchyVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, scale, name, learnable=False, is_observed=False):
        self._type = "HalfCauchy"
        ranges = {"scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.HalfCauchyDistribution()


class HalfNormalVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, scale, name, learnable=False, is_observed=False):
        self._type = "HalfNormal"
        ranges = {"scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.HalfNormalDistribution()


class Chi2Variable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, df, name, learnable=False, is_observed=False):
        self._type = "Chi2"
        ranges = {"df": geometric_ranges.UnboundedRange()} #TODO: Natural number
        super().__init__(name, df=df, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.Chi2Distribution()


class GumbelVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, loc, scale, name, learnable=False, is_observed=False):
        self._type = "Gumbel"
        ranges = {"loc": geometric_ranges.UnboundedRange(),
                  "scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.GumbelDistribution()


class LaplaceVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, loc, scale, name, learnable=False, is_observed=False):
        self._type = "Laplace"
        ranges = {"loc": geometric_ranges.UnboundedRange(),
                  "scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.LaplaceDistribution()


class ExponentialVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, rate, name, learnable=False, is_observed=False):
        self._type = "Exponential"
        ranges = {"rate": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, rate=rate, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.ExponentialDistribution()


class PoissonVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, rate, name, learnable=False, is_observed=False):
        self._type = "Poisson"
        ranges = {"rate": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, rate=rate, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.PoissonDistribution()


class LogNormalVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, loc, scale, name, learnable=False, is_observed=False):
        self._type = "Log Normal"
        ranges = {"loc": geometric_ranges.UnboundedRange(),
                  "scale": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.LogNormalDistribution()


#class LogitNormalVariable(VariableConstructor):
#    """
#    Summary
#
#    Parameters
#    ----------
#    """
#    def __init__(self, loc, scale, name, learnable=False, is_observed=False):
#        self._type = "Logit Normal"
#        ranges = {"loc": geometric_ranges.UnboundedRange(),
#                  "scale": geometric_ranges.RightHalfLine(0.)}
#        super().__init__(name, loc=loc, scale=scale, learnable=learnable, ranges=ranges, is_observed=is_observed)
#        self.distribution = distributions.LogitNormalDistribution()


class BetaVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, alpha, beta, name, learnable=False, is_observed=False):
        self._type = "Beta"
        concentration1 = alpha
        concentration0 = beta
        ranges = {"concentration1": geometric_ranges.RightHalfLine(0.),
                  "concentration0": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, concentration1=concentration1, concentration0=concentration0,
                         learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.BetaDistribution()


class BinomialVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, total_count, probs=None, logits=None, name="Binomial", learnable=False, is_observed=False):
        self._type = "Binomial"
        if probs is not None and logits is None:
            ranges = {"total_count": geometric_ranges.UnboundedRange(), #TODO: It should become natural number in the future
                      "probs": geometric_ranges.Interval(0., 1.)}
            super().__init__(name, total_count=total_count, probs=probs, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.BinomialDistribution()
        elif logits is not None and probs is None:
            ranges = {"total_count": geometric_ranges.UnboundedRange(),
                      "logits": geometric_ranges.UnboundedRange()}
            super().__init__(name, total_count=total_count, logits=logits, learnable=learnable, ranges=ranges)
            self.distribution = distributions.BinomialDistribution()
        else:
            raise ValueError("Either probs or " +
                             "logits needs to be provided as input")


class NegativeBinomialVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, total_count, probs=None, logits=None, name="NegativeBinomial", learnable=False, is_observed=False):
        self._type = "NegativeBinomial"
        if probs is not None and logits is None:
            ranges = {"total_count": geometric_ranges.UnboundedRange(), #TODO: It should become natural number in the future
                      "probs": geometric_ranges.Interval(0., 1.)}
            super().__init__(name, total_count=total_count, probs=probs, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.BinomialDistribution()
        elif logits is not None and probs is None:
            ranges = {"total_count": geometric_ranges.UnboundedRange(),
                      "logits": geometric_ranges.UnboundedRange()}
            super().__init__(name, total_count=total_count, logits=logits, learnable=learnable, ranges=ranges)
            self.distribution = distributions.NegativeBinomialDistribution()
        else:
            raise ValueError("Either probs or " +
                             "logits needs to be provided as input")


class BernoulliVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, probs=None, logits=None, name="Bernoulli", learnable=False, is_observed=False):
        self._type = "Bernoulli"
        if probs is not None and logits is None:
            ranges = {"probs": geometric_ranges.Interval(0., 1.)}
            super().__init__(name, probs=probs, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.BernoulliDistribution()
        elif logits is not None and probs is None:
            ranges = {"logits": geometric_ranges.UnboundedRange()}
            super().__init__(name, logits=logits, learnable=learnable, ranges=ranges)
            self.distribution = distributions.BernoulliDistribution()
        else:
            raise ValueError("Either probs or " +
                             "logits needs to be provided as input")


class GeometricVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, probs=None, logits=None, name="Geometric", learnable=False, is_observed=False):
        self._type = "Geometric"
        if probs is not None and logits is None:
            ranges = {"probs": geometric_ranges.Interval(0., 1.)}
            super().__init__(name, probs=probs, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.GeometricDistribution()
        elif logits is not None and probs is None:
            ranges = {"logits": geometric_ranges.UnboundedRange()}
            super().__init__(name, logits=logits, learnable=learnable, ranges=ranges)
            self.distribution = distributions.GeometricDistribution()
        else:
            raise ValueError("Either probs or " +
                             "logits needs to be provided as input")


class CategoricalVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, probs=None, logits=None, name="Categorical", learnable=False, is_observed=False):
        self._type = "Categorical"
        if probs is not None and logits is None:
            ranges = {"p": geometric_ranges.Simplex()}
            super().__init__(name, probs=probs, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.CategoricalDistribution()
        elif logits is not None and probs is None:
            ranges = {"logits": geometric_ranges.UnboundedRange()}
            super().__init__(name, logits=logits, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.CategoricalDistribution()
        else:
            raise ValueError("Either probs or " +
                             "logits needs to be provided as input")


#class ConcreteVariable(VariableConstructor):
#    """
#    Summary
#
#    Parameters
#    ----------
#    """
#    def __init__(self, tau, probs, name, learnable=False, is_observed=False):
#        self._type = "Concrete"
#        ranges = {"tau": geometric_ranges.RightHalfLine(0.),
#                  "p": geometric_ranges.Simplex()}
#        super().__init__(name, tau=tau, probs=probs, learnable=learnable, ranges=ranges, is_observed=is_observed)
#        self.distribution = distributions.ConcreteDistribution()


class MultivariateNormalVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, loc, covariance_matrix=None, precision_matrix=None,
                 scale_tril=None, name="Multivariate Normal", learnable=False, is_observed=False):
        self._type = "Multivariate Normal"
        if scale_tril is not None and covariance_matrix is None and precision_matrix is None:
            ranges = {"loc": geometric_ranges.UnboundedRange(),
                      "scale_tril": geometric_ranges.UnboundedRange()}
            super().__init__(name, loc=loc, scale_tril=scale_tril, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.MultivariateNormalDistribution()

        elif scale_tril is None and covariance_matrix is not None and precision_matrix is None:
            ranges = {"loc": geometric_ranges.UnboundedRange(),
                      "covariance_matrix": geometric_ranges.PositiveDefiniteMatrix()}
            super().__init__(name, loc=loc, covariance_matrix=covariance_matrix, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.MultivariateNormalDistribution()

        elif scale_tril is None and covariance_matrix is None and precision_matrix is not None:
            ranges = {"loc": geometric_ranges.UnboundedRange(),
                      "precision_matrix": geometric_ranges.UnboundedRange()}
            super().__init__(name, loc=loc, precision_matrix=precision_matrix, learnable=learnable, ranges=ranges, is_observed=is_observed)
            self.distribution = distributions.MultivariateNormalDistribution()

        else:
            raise ValueError("Either covariance_matrix or precision_matrix or"+
                             "scale_tril needs to be provided as input")


class DirichletVariable(VariableConstructor):
    """
    Summary

    Parameters
    ----------
    """
    def __init__(self, concentration, name, learnable=False, is_observed=False):
        self._type = "Dirichlet"
        ranges = {"concentration": geometric_ranges.RightHalfLine(0.)}
        super().__init__(name, concentration=concentration,
                         learnable=learnable, ranges=ranges, is_observed=is_observed)
        self.distribution = distributions.DirichletDistribution()


"""
NN.modules
---------
Module description
"""

from torch import nn

class ParameterModule(nn.Module):
    """
    Summary
    """
    def __init__(self, parameter):
        super(ParameterModule, self).__init__()
        self.parameter = parameter

    def __call__(self, *args, **kwargs):
        return self.parameter

class EmptyModule(nn.ModuleList):
    """
    Summary
    """
    def __init__(self):
        links = []
        super(EmptyModule, self).__init__(links)

